import argparse
import datetime
import logging
import os
import re
import shutil
import sys
from copy import deepcopy

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

import rck
from rck.utils.adj.process import refined_adjacencies_reciprocal
from rck.core.io import read_adjacencies_from_file, write_acnt_to_file, get_logging_cli_parser, get_standard_logger_from_args, read_adjacency_groups_from_file, \
    read_segments_from_file, \
    extract_scnt_from_segments, extract_scnb_from_segments, read_scnb_from_file, get_full_path, write_scnt_to_file, EXTERNAL_NA_ID, FALSE_POSITIVE, remove_cn_data_from_segments, \
    write_scnb_to_file, write_adjacencies_to_file, write_segments_to_file, remove_cn_data_from_adjacencies, read_positions_from_file, \
    write_positions_to_file, remove_cnb_data_from_segments
from rck.core.process import positions_aligned, adj_groups_concur
from rck.core.structures import get_segments_for_fragments_ids_dict, get_ref_telomeres_from_segments, get_ref_adjacencies_from_segments, SegmentCopyNumberBoundaries, refined_scnt, \
    refined_scnb, refined_scnt_with_adjacencies_and_telomeres, extract_spanned_extremities, SCNBoundariesStrategies, LengthSpreadRelationships, AdjacencyType, AdjacencyGroupType, \
    Phasing, Haplotype, CNBoundaries, AdjacencyCopyNumberProfile, SegmentCopyNumberProfile
from rck.core.graph import construct_hiag_inflate_from_haploid_data, IntervalAdjacencyGraph


def main():
    parser = argparse.ArgumentParser(parents=[get_logging_cli_parser()])
    parser.add_argument('--version', action='version', version=rck.version)

    parser.add_argument("--scnt", required=True)
    parser.add_argument("--scnt-separator", default="\t")
    parser.add_argument("--scnt-extra-separator", default=";")

    parser.add_argument("--adjacencies", required=True)
    parser.add_argument("--adjacencies-separator", default="\t")
    parser.add_argument("--adjacencies-extra-separator", default=";")

    parser.add_argument("--workdir", default=None)

    parser.add_argument("--clone-ids", default=None)

    parser.add_argument("--scnb", default=None)
    parser.add_argument("--scnb-separator", default="\t")
    parser.add_argument("--scnb-extra-separator", default=";")

    parser.add_argument("--telomere-segments", default=None)
    parser.add_argument("--telomere-segments-separator", default="\t")
    parser.add_argument("--telomere-segments-extra-separator", default=";")

    parser.add_argument("--telomere-positions", default=None)
    parser.add_argument("--telomere-positions-separator", default="\t")
    parser.add_argument("--telomere-positions-no-reciprocal", action="store_false", dest="telomere_reciprocal")

    parser.add_argument("--adjacency-groups", default=None)
    parser.add_argument("--adjacency-group-separator", default="\t")
    parser.add_argument("--adjacency-group-extra-separator", default=";")

    parser.add_argument("--fragments", default=None)
    parser.add_argument("--fragments-separator", default="\t")
    parser.add_argument("--fragments-extra-separator", default=";")

    # parser.add_argument("--trees", default=None)
    parser.add_argument("--no-allow-unit-segments", action="store_false", dest="allow_unit_segments")
    ###
    pre_group = parser.add_argument_group()
    pre_group.add_argument("--no-pre", action="store_false", dest="do_preprocess")
    ##
    # Arguments for Segment Copy Number Tensor preprocessing
    ##
    pre_group.add_argument("--pre-no-scnt", action="store_false", dest="do_pre_scnt")
    pre_group.add_argument("--pre-scnt-no-merge-fragments", action="store_false", dest="pre_scnt_merge_fragments")
    pre_group.add_argument("--pre-scnt-max-merge-gap", type=int, default=1000000000)
    pre_group.add_argument("--pre-scnt-no-fill-gaps", action="store_false", dest="pre_scnt_fill_gaps")
    pre_group.add_argument("--pre-scnt-max-fill-gap", type=int, default=1000000000)
    ##
    # Arguments for Segment Copy Number Boundaries
    ##
    pre_group.add_argument("--pre-scnb", action="store_true", dest="do_pre_bnd")
    pre_group.add_argument("--pre-scnb-strategy", choices=[strategy.value for strategy in SCNBoundariesStrategies], type=SCNBoundariesStrategies.from_string,
                           default=SCNBoundariesStrategies.UNIFORM_MIN_MAX.value)
    pre_group.add_argument("--pre-scnb-uniform-spread-size", type=int, default=1)
    pre_group.add_argument("--pre-scnb-length-spread-relation", choices=[rel.value for rel in LengthSpreadRelationships], type=LengthSpreadRelationships.from_string,
                           default=LengthSpreadRelationships.DUMMY.value)
    pre_group.add_argument("--pre-scnb-uniform-min", type=int, default=0)
    pre_group.add_argument("--pre-scnb-uniform-max", type=int, default=10)
    pre_group.add_argument("--pre-scnb-min-allow-zero-for-positive", type=int, default=-1)
    pre_group.add_argument("--pre-scnb-max-allow-zero-for-positive", type=int, default=1000000000)
    pre_group.add_argument("--pre-scnb-min-allow-positive-for-zero", type=int, default=-1)
    pre_group.add_argument("--pre-scnb-max-allow-positive-for-zero", type=int, default=1000000000)
    pre_group.add_argument("--pre-scnb-is-male", action="store_false", dest="pre_scnb_is_female")
    ##
    # Arguments for adjacencies
    ##
    pre_group.add_argument("--pre-no-adj", action="store_false", dest="do_pre_adj")
    pre_group.add_argument("--pre-adj-no-reciprocal", action="store_false", dest="pre_adj_reciprocal")
    pre_group.add_argument("--pre-adj-reciprocal-include-ref", action="store_true", dest="pre_adj_reciprocal_include_ref")
    pre_group.add_argument("--pre-adj-reciprocal-max-distance", type=int, default=50)
    ###
    run_group = parser.add_argument_group()
    run_group.add_argument("--no-run", action="store_false", dest="do_run")
    run_group.add_argument("--run-g-mip-gap", type=float, default=0.015)
    run_group.add_argument("--run-g-time-limit", type=int, default=28800)
    run_group.add_argument("--run-g-threads", type=int, default=4)
    run_group.add_argument("--run-g-mip-focus", type=int, choices=[0, 1, 2, 3], default=0)
    run_group.add_argument("--run-nas-fp", type=float, default=0.1)
    run_group.add_argument("--run-group-m-default-fp", type=float, default=0.1)
    run_group.add_argument("--run-group-n-default-fp", type=float, default=0.1)
    run_group.add_argument("--run-segment-length-attr", choices=["length", "length_10", "length_100", "length_1000"], default="length_10")
    run_group.add_argument("--run-g-allow-interrupted", action="store_true")
    ###
    output_group = parser.add_argument_group()
    output_group.add_argument("--o-prefix-name", type=str, dest="out_prefix_name", default="")
    output_group.add_argument("--o-scnt-separator", default="\t")
    output_group.add_argument("--o-scnt-extra-separator", default=";")
    output_group.add_argument("--o-acnt-separator", default="\t")
    output_group.add_argument("--o-acnt-extra-separator", default=";")
    output_group.add_argument("--o-acnt-mix-novel-and-reference", action="store_true",)
    ###
    post_group = parser.add_argument_group()
    post_group.add_argument("--post-check-all", action="store_true", dest="post_check_all")
    post_group.add_argument("--post-check-scnb", action="store_true", dest="post_check_scnb")
    post_group.add_argument("--post-check-labeling", action="store_true", dest="post_check_labeling")
    post_group.add_argument("--post-check-balancing", action="store_true", dest="post_check_balancing")
    post_group.add_argument("--post-check-adj-groups", action="store_true", dest="post_check_adj_groups")
    post_group.add_argument("--post-check-adj-groups-m", action="store_true", dest="post_check_adj_groups_m")
    post_group.add_argument("--post-check-adj-groups-n", action="store_true", dest="post_check_adj_groups_n")
    post_group.add_argument("--post-check-nas-fp", action="store_true", dest="post_check_nas_fp")
    ###
    args = parser.parse_args()

    from rck.core.ilp_gurobi import OptModelMultiClone, DEFAULT_GROUP_M_FP, SEGMENT_LENGTH_ATTRIBUTE
    import gurobi as g
    logger = get_standard_logger_from_args(args=args, program_name="RCK")

    workdir = args.workdir if args.workdir is not None else "RCK-{date}".format(date=str(datetime.date.today()))
    workdir_path = get_full_path(path=workdir)
    logger.debug("Working directory is set as {workdir_path}".format(workdir_path=workdir_path))

    logger.debug("Creating working directory if does not exist")
    os.makedirs(workdir_path, exist_ok=True)
    internal_debug_log_file = os.path.join(workdir_path, "debug.log")
    formatter = logging.Formatter(args.log_format)
    fh = logging.FileHandler(internal_debug_log_file)
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    raw_input_dir_path = get_full_path(path=os.path.join(workdir_path, "raw_input"))
    logger.debug("Raw input directory is {raw_dir_path}".format(raw_dir_path=raw_input_dir_path))

    logger.debug("Creating raw input directory if does not exist")
    os.makedirs(raw_input_dir_path, exist_ok=True)

    scnt_file_path = get_full_path(path=args.scnt)
    logger.debug("Allele-specific segment copy number tensor file path {scnt_path}".format(scnt_path=scnt_file_path))
    logger.debug("Copying SCNT file {scnt_file} to raw input direcotry {raw_input_dir}".format(scnt_file=scnt_file_path, raw_input_dir=raw_input_dir_path))
    shutil.copy2(src=scnt_file_path, dst=raw_input_dir_path)
    logger.info("Reading allele-specific segment copy number tensor from {scnt_path}".format(scnt_path=scnt_file_path))
    segments = read_segments_from_file(file_name=scnt_file_path, separator=args.scnt_separator, extra_separator=args.scnt_extra_separator)
    try:
        logger.info("Trying to extract allele-specific SCNT from segment copy number tensor from {scnt_path}".format(scnt_path=scnt_file_path))
        scnt = extract_scnt_from_segments(segments=segments)
    except ValueError as e:
        logger.error("Unable to get the allele-specific segment copy number information from {scnt_path}. Terminating".format(scnt_path=scnt_file_path))
        raise e

    clone_ids = sorted(set(args.clone_ids.split(","))) if args.clone_ids is not None else sorted(scnt.keys())
    logger.info("Clones are set as {clone_ids}".format(clone_ids=",".join(clone_ids)))
    logger.debug("Clone ids in allele-specific segment copy number tensor are {clone_ids}.".format(clone_ids=sorted(scnt.keys())))
    if set(clone_ids) < set(scnt.keys()):
        logger.info("Segment copy number tensor has information about clones {scnt_clones}, while only clones {clone_ids} were specified for processing."
                    "".format(scnt_clones=sorted(scnt.keys()), clone_ids=clone_ids))
    elif set(clone_ids) > set(scnt.keys()):
        logger.error("Segment copy number tensor has information about clones {scnt_clones}, but only clones {clone_ids} were specified for processing. Error."
                     "".format(scnt_clones=sorted(scnt.keys()), clone_ids=clone_ids))
    new_scnt = {}
    for clone_id in clone_ids:
        if clone_id not in scnt:
            raise ValueError("Clone {clone_id} is not present for allele-specific segment copy number values".format(clone_id=clone_id))
        new_scnt[clone_id] = scnt[clone_id]
    scnt = new_scnt

    try:
        logger.info("Trying to extract segment copy number boundaries from {scnt_path}".format(scnt_path=scnt_file_path))
        scnb = extract_scnb_from_segments(segments=segments, clone_ids=clone_ids, allow_missing=True)
        logger.info("Successfully extracted (at least some) segment copy number boundaries from {scnt_path}".format(scnt_path=scnt_file_path))
    except ValueError:
        scnb = {clone_id: SegmentCopyNumberBoundaries() for clone_id in clone_ids}
        logger.info("Unsuccessfully tried to extract segment copy number boundaries from {scnt_path}. Not to worry just yet.".format(scnt_path=scnt_file_path))

    scnb_file_path = None if args.scnb is None else get_full_path(args.scnb)
    try:
        if scnb_file_path is not None:
            logger.debug("Copying Segment Copy Number Boundaries file {cnb_file} to raw input directory {raw_input_dir}".format(cnb_file=scnb_file_path,
                                                                                                                                raw_input_dir=raw_input_dir_path))
            shutil.copy2(src=scnb_file_path, dst=raw_input_dir_path)
            logger.info("Trying to read segment copy number boundaries data from file {scnb_path}.".format(scnb_path=scnb_file_path))
            segments_from_scnb_file, scnb_from_file = read_scnb_from_file(file_name=scnt_file_path, clone_ids=clone_ids, allow_missing=True,
                                                                          separator=args.scnb_separator, extra_separator=args.scnb_extra_separator)
            logger.info("Successfully extracted (at least some) segment copy numbers from {scnb_path}. Updating previous segment copy number boundaries data (if exists)."
                        "".format(scnb_path=scnb_from_file))
            for clone_id in clone_ids:
                scnb[clone_id].update(scnb_from_file[clone_id])
    except ValueError:
        logger.info("Unsuccessfully tried to extract segment copy number boundaries from {scnb_path}. Not to worry just yet.".format(scnb_path=scnb_file_path))

    if args.do_preprocess and args.do_pre_scnt:
        new_segments, scnt, segments_ids_mapping = refined_scnt(segments=segments, scnt=scnt,
                                                                merge_fragments=args.pre_scnt_merge_fragments, max_merge_gap=args.pre_scnt_max_merge_gap,
                                                                fill_gaps=args.pre_scnt_fill_gaps, max_fill_gap=args.pre_scnt_max_fill_gap)
        old_segment, segments = segments, new_segments
        scnb = refined_scnb(scnb=scnb, new_segments=segments, segments_ids_mapping=segments_ids_mapping)

    adjacencies_file_path = get_full_path(path=args.adjacencies)
    logger.debug("Adjacencies file path {file_path}".format(file_path=adjacencies_file_path))

    logger.debug("Copying NAs file {nas_file} to raw input directory {raw_input_dir}".format(nas_file=adjacencies_file_path, raw_input_dir=raw_input_dir_path))
    shutil.copy2(src=adjacencies_file_path, dst=raw_input_dir_path)

    logger.info("Reading Adjacencies from {file_path}".format(file_path=adjacencies_file_path))
    input_adjacencies = read_adjacencies_from_file(file_name=adjacencies_file_path, separator=args.adjacencies_separator, extra_separator=args.adjacencies_extra_separator)
    logger.debug("A total of {cnt} unlabeled novel adjacencies were obtained".format(cnt=len([a for a in input_adjacencies if a.adjacency_type == AdjacencyType.NOVEL])))
    logger.debug("A total of {cnt} unlabeled reference adjacencies were obtained".format(cnt=len([a for a in input_adjacencies if a.adjacency_type == AdjacencyType.REFERENCE])))
    if args.do_preprocess and args.do_pre_adj:
        logger.info("Preprocessing input adjacencies.")
        if args.pre_adj_reciprocal:
            input_ref = [adj for adj in input_adjacencies if adj.adjacency_type == AdjacencyType.REFERENCE]
            input_nov = [adj for adj in input_adjacencies if adj.adjacency_type == AdjacencyType.NOVEL]
            adjacencies = input_nov
            if args.pre_adj_reciprocal_include_ref:
                logger.debug("Both novel and reference input adjacencies are preprocessed w.r.t. reciprocality")
                adjacencies = adjacencies + input_ref
            logger.info("Preprocessing input adjacencies w.r.t. reciprocality.")
            adjacencies = refined_adjacencies_reciprocal(novel_adjacencies=adjacencies, max_distance=args.pre_adj_reciprocal_max_distance)
            if args.pre_adj_reciprocal_include_ref:
                input_adjacencies = adjacencies
            else:
                input_adjacencies = adjacencies + input_ref

    telomere_segments_file_path = None if args.telomere_segments is None else get_full_path(path=args.telomere_segments)
    logger.debug("Additional telomeres file path {tel_path}".format(tel_path=telomere_segments_file_path))
    if telomere_segments_file_path is not None:
        logger.debug("Copying telomere segments file {telomeres_file} to raw input directory {raw_input_dir}".format(telomeres_file=telomere_segments_file_path,
                                                                                                                     raw_input_dir=raw_input_dir_path))
        shutil.copy2(src=telomere_segments_file_path, dst=raw_input_dir_path)
        logger.info("Reading additional telomere locations (via segments) from {tel_path}".format(tel_path=telomere_segments_file_path))
        telomeres_segments = read_segments_from_file(file_name=telomere_segments_file_path, separator=args.telomere_segments_separator,
                                                     extra_separator=args.telomere_segments_extra_separator)
    else:
        telomeres_segments = []
        logger.info("No additional telomeres positions (via segments) were provided, only reference locations are allowed.")

    telomere_positions_file_path = None if args.telomere_positions is None else get_full_path(path=args.telomere_positions)
    if telomere_positions_file_path is not None:
        logger.debug("Copying telomere positions file {telomeres_file} to raw input directory {raw_input_dir}".format(telomeres_file=telomere_positions_file_path,
                                                                                                                      raw_input_dir=raw_input_dir_path))
        shutil.copy2(src=telomere_positions_file_path, dst=raw_input_dir_path)
        logger.info("Reading additional telomere locations (via positions) from {tel_path}".format(tel_path=telomere_positions_file_path))
        input_telomere_positions = read_positions_from_file(file_name=telomere_positions_file_path, separator=args.telomere_positions_separator)
    else:
        logger.info("No additional telomeres positions (via positions) were provided, only reference locations are allowed.")
        input_telomere_positions = []

    adjacency_group_file_path = None if args.adjacency_groups in ("", None) else get_full_path(path=args.adjacency_groups)
    logger.debug("Adjacency groups file path {nas_groups_file}".format(nas_groups_file=adjacency_group_file_path))
    if adjacency_group_file_path is not None:
        logger.info("Reading adjacencies groups from {nas_groups_file}".format(nas_groups_file=adjacency_group_file_path))
        adjacency_groups = read_adjacency_groups_from_file(file_name=adjacency_group_file_path)
    else:
        adjacency_groups = []
        logger.info("No information about adjacency groups was provided")

    fragments_file_path = None if args.fragments in (None, "") else get_full_path(path=args.fragments)
    if fragments_file_path is not None:
        logger.debug("Copying fragments file {fragments_file} to raw input directory {raw_input_dir}".format(fragments_file=fragments_file_path,
                                                                                                             raw_input_dir=raw_input_dir_path))
        shutil.copy2(src=fragments_file_path, dst=raw_input_dir_path)
        fragments = read_segments_from_file(file_name=fragments_file_path)
    else:
        fragments = None
        logger.debug("Fragments file path {fragments_file_path}".format(fragments_file_path=fragments_file_path))

    overall_nas_fp = args.run_nas_fp
    logger.info("Global False Positive value for novel adjacencies is set at {fp}".format(fp=overall_nas_fp))

    preprocessed_input_dir_path = get_full_path(os.path.join(workdir_path, "input"))
    os.makedirs(preprocessed_input_dir_path, exist_ok=True)
    if fragments is None:
        fragments = deepcopy(segments)

    if args.do_preprocess:
        logger.info("Refining input adjacencies and segments/fragments, so that extremities are aligned.")
        fragments = deepcopy(segments)
        segments, scnt, segments_ids_mapping = refined_scnt_with_adjacencies_and_telomeres(segments=segments, scnt=scnt, adjacencies=input_adjacencies,
                                                                                           telomere_positions=input_telomere_positions)
        scnb = refined_scnb(scnb=scnb, new_segments=segments, segments_ids_mapping=segments_ids_mapping)

    logger.info("Creating missing segment copy number boundaries will be automatically generated.")
    for clone_id in clone_ids:
        missing_only = not args.do_pre_bnd
        scnb[clone_id].fill(segments=segments, scnp=scnt[clone_id], missing_only=missing_only, strategy=args.pre_scnb_strategy,
                            min_allow_zero_for_positive=args.pre_scnb_min_allow_zero_for_positive,
                            max_allow_zero_for_positive=args.pre_scnb_max_allow_zero_for_positive,
                            min_allow_positive_for_zero=args.pre_scnb_min_allow_positive_for_zero,
                            max_allow_positive_for_zero=args.pre_scnb_max_allow_positive_for_zero,
                            uniform_spread_size=args.pre_scnb_uniform_spread_size,
                            length_spread_relation=args.pre_scnb_length_spread_relation,
                            uniform_min=args.pre_scnb_uniform_min,
                            uniform_max=args.pre_scnb_uniform_max,
                            is_female=args.pre_scnb_is_female)

    if telomeres_segments is not None:
        logger.info("Extracting additional permitted telomere locations")
        additional_telomeres = extract_spanned_extremities(source=segments, boundaries=telomeres_segments)
        logger.debug("Extracted {cnt} additional telomere locations".format(cnt=len(additional_telomeres)))
        input_telomere_positions.extend(additional_telomeres)

    logger.info("A total of input telomere locations is {cnt}".format(cnt=len(input_telomere_positions)))

    logger.info("Finished preprocessing. At this point every data part must be aligned with each other.")
    logger.info("Checking data alignment/concordance")
    if fragments is None:
        logger.error("FRAGMENTS must exist; their absence at this point indicates that preprocessing was turned off, while --fragments were not part of the original input")
        exit(1)
    if scnb is None:
        logger.error("Lower/Upper boundaries for Segment Copy Numbers must exist; "
                     "their absence at this point indicated that preprocessing was turned off, while --scn-boundaries were not part of the original input")
        exit(1)
    segments_to_fragments = None
    try:
        segments_to_fragments = get_segments_for_fragments_ids_dict(segments=segments, fragments=fragments, allow_non_covered=False)
    except Exception:
        logger.error("FRAGMENTS boundaries do not align with SEGMENTS boundaries; "
                     "having this problem at this point indicates that the preprocessing was turned off and the data in original input SCNT and --fragments is not aligned")
        exit(1)
    segments_positions = []
    for s in segments:
        segments_positions.append(s.start_position)
        segments_positions.append(s.end_position)
    nas_positions = []
    for na in input_adjacencies:
        nas_positions.append(na.position1)
        nas_positions.append(na.position2)
    logger.debug("Checking novel adjacency positions concordance")
    if not positions_aligned(segments_positions=segments_positions, other_positions=nas_positions):
        logger.error("NOVEL ADJACENCIES positions are not aligned with extremities of SEGMENTS; "
                     "having this problem at this point indicates that the preprocessing was turned off and the data in original input SCNT and NAS is not aligned")
        exit(1)
    telomeres = sorted(get_ref_telomeres_from_segments(segments=segments) + input_telomere_positions, key=lambda p: p.coordinate)
    logger.debug("Checking telomere positions concordance")
    if not positions_aligned(segments_positions=segments_positions, other_positions=telomeres):
        logger.error("TELOMERES positions are not aligned with extremities of SEGMENTS; "
                     "having this problem at this problem indicates that preprocessing was turned off and the data in the original SCNT and --telomeres is not aligned")
        exit(1)
    logger.debug("Checking adjacency groups concordance")
    if len(adjacency_groups) > 0 and not adj_groups_concur(adj_groups=adjacency_groups, adjacencies=input_adjacencies):
        logger.error("Some novel adjacencies group in NAS-GROUPS refers to an absent ADJACENCY; "
                     "having this problem at this time indicates that either the original --nas-groups has a group that is referring to the missing novel adjacency in NAS, "
                     "or the referred novel adjacency from NAS was removed during the preprocessing.")
        exit(1)
    logger.debug("Populating adjacency groups")
    adjacencies_by_external_ids = {adj.extra.get(EXTERNAL_NA_ID, adj.stable_id_non_phased): adj for adj in input_adjacencies}
    for ag in adjacency_groups:
        ag.populate_adjacencies_via_ids(source=input_adjacencies, source_by_ids=adjacencies_by_external_ids)

    ########
    #
    # Creating files with data, that can be used without running preprocessing in the future
    #
    ########

    preprocessed_scnt_file = os.path.join(preprocessed_input_dir_path, "rck.scnt.tsv")
    remove_cn_data_from_segments(segments=segments)
    remove_cnb_data_from_segments(segments=segments)
    logger.info("Writing (preprocessed) allele-specific segment copy number data to {file}".format(file=preprocessed_scnt_file))
    write_scnt_to_file(file_name=preprocessed_scnt_file, segments=segments, scnt=scnt, clone_ids=clone_ids, inplace=False)

    preprocessed_scnb_file = os.path.join(preprocessed_input_dir_path, "rck.scnb.tsv")
    logger.info("Writing (preprocessed) segment copy number boundaries information to {file}".format(file=preprocessed_scnb_file))
    write_scnb_to_file(file_name=preprocessed_scnb_file, segments=segments, scnb=scnb, clone_ids=clone_ids, inplace=False)

    preprocessed_input_adjacencies_file = os.path.join(preprocessed_input_dir_path, "rck.adj.tsv")
    logger.info("Writing (preprocessed) input adjacencies data to {file}".format(file=preprocessed_input_adjacencies_file))
    write_adjacencies_to_file(file_name=preprocessed_input_adjacencies_file, adjacencies=input_adjacencies)
    remove_cn_data_from_adjacencies(input_adjacencies)

    preprocessed_fragments_file = os.path.join(preprocessed_input_dir_path, "rck.frag.tsv")
    remove_cn_data_from_segments(segments=fragments)
    logger.info("Writing (preprocessed) fragments data to {file}".format(file=preprocessed_fragments_file))
    write_segments_to_file(file_name=preprocessed_fragments_file, segments=fragments, extra=None)

    ########

    logger.debug("Extracting reference adjacencies from input segments")
    ref_adjacencies = get_ref_adjacencies_from_segments(segments=segments, assign_external_ids=True)
    logger.debug("A total of {cnt} reference unlabeled adjacencies are considered (labeled will be x2, except for X and Y chromosomes)".format(cnt=len(ref_adjacencies)))
    adjacencies = input_adjacencies + ref_adjacencies
    output_dir = os.path.join(workdir_path, "output")
    logger.debug("Output directory is {output}".format(output=output_dir))
    os.makedirs(output_dir, exist_ok=True)

    if args.telomere_reciprocal:
        logger.info("Reciprocal telomere positions are specified via --telomere-positions-reciprocal. Extracting.")
        iag = IntervalAdjacencyGraph(segments=segments, adjacencies=adjacencies)
        iag.build_graph()
        for u, v in iag.ref_adjacency_edges(data=False):
            if u in telomeres and v not in telomeres:
                telomeres.append(v)
            if v in telomeres and u not in telomeres:
                telomeres.append(u)
    preprocessed_telomeres_file = os.path.join(preprocessed_input_dir_path, "rck.pos.tsv")
    logger.info("Writing (preprocessed) telomeres (both input and reference) to {file}".format(file=preprocessed_telomeres_file))
    write_positions_to_file(file_name=preprocessed_telomeres_file, positions=telomeres)

    logger.debug("Output directory is {output}".format(output=output_dir))
    os.makedirs(output_dir, exist_ok=True)

    ##########
    #
    # Gurobi model
    #
    ##########
    if not args.do_run:
        logger.info("A --no-run flag was set. Not performing the inference.")
        exit(0)

    gurobi_log_path = os.path.join(output_dir, "gurobi.log")
    logger.info("Gurobi log will be stored in {log_path}".format(log_path=gurobi_log_path))
    extra = {DEFAULT_GROUP_M_FP: args.run_group_m_default_fp,
             SEGMENT_LENGTH_ATTRIBUTE: args.run_segment_length_attr}
    logger.debug("Gurobi ILP extra is {extra}".format(extra=str(extra)))
    logger.info("Setting up Gurobi ILP model (includes construction of the IAG)")
    ilp_model = OptModelMultiClone(hapl_segments=segments,
                                   hapl_adjacencies=adjacencies,
                                   scnt=scnt,
                                   hapl_telomeres=telomeres,
                                   hapl_segments_to_fragments=segments_to_fragments,
                                   hapl_adjacencies_groups=adjacency_groups,
                                   scnb=scnb,
                                   hapl_nov_adjacencies_fp=overall_nas_fp,
                                   extra=extra)
    logger.debug("Building variables and constraints")
    ilp_model.build_gurobi_model()
    logger.debug("Setting gurobi parameters")
    ilp_model.gm.setParam("MIPGap", args.run_g_mip_gap)
    ilp_model.gm.setParam("MIPGapAbs", args.run_g_mip_gap)
    ilp_model.gm.setParam("MIPFocus", args.run_g_mip_focus)
    ilp_model.gm.setParam("LogFile", gurobi_log_path)
    ilp_model.gm.setParam("TimeLimit", args.run_g_time_limit)
    ilp_model.gm.setParam("Threads", args.run_g_threads)
    logger.info("Starting Gurobi to solve the optimization problem")
    ilp_model.solve_model()

    logger.info("Gurobi model solving has ended")
    status = ilp_model.gm.status
    solution_cnt = ilp_model.gm.solcount
    if status == g.GRB.Status.INFEASIBLE:
        logger.error("Constructed model was infeasible to solve. This usually happens because of the tight segment copy number boundaries.")
        logger.error("\t Try more flexible segment copy number boundaries or try increasing allowed Novel adjacencies False Positive value.")
        logger.info("Gurobi computes the IIS")
        ilp_model.gm.computeIIS()
        ilp_path = os.path.join(output_dir, "model.ilp")
        logger.info("Writing down Gurobi IIS to {file}".format(file=ilp_path))
        ilp_model.gm.write(ilp_path)
        logger.error("Inference was unsuccessful.")
        exit(1)
    allowed_statuses = [g.GRB.Status.OPTIMAL, g.GRB.Status.TIME_LIMIT]
    if args.run_g_allow_interrupted:
        allowed_statuses.append(g.GRB.Status.INTERRUPTED)
    if status not in allowed_statuses:
        logger.error("Gurobi finished with status {status}".format(status=status))
        logger.error("Inference was unsuccessful")
        exit(1)
    if solution_cnt == 0:
        logger.error("Inference was unsuccessful")
        exit(1)

    logger.info("Gurobi finished execution with status {status}".format(status=status))
    logger.info("Extracting inferred diploid segment copy number data")
    scnt = ilp_model.get_scnt_from_model()
    logger.info("Extracting inferred diploid adjacency copy number data")
    acnt = ilp_model.get_acnt_from_model()
    scnt_base_name = args.out_prefix_name + "rck.scnt.tsv"
    if scnt_base_name.startswith("."):
        scnt_base_name = scnt_base_name[1:]
    scnt_base_name = re.sub("\.+", ".", scnt_base_name)
    scnt_file_path = os.path.join(output_dir, scnt_base_name)
    logger.info("Writing inferred diploid segment copy number data to {file}".format(file=scnt_file_path))
    write_scnt_to_file(file_name=scnt_file_path, segments=segments, scnt=scnt)
    acnt_base_name = args.out_prefix_name + "rck.acnt.tsv"
    if acnt_base_name.startswith("."):
        acnt_base_name = acnt_base_name[1:]
    acnt_base_name = re.sub("\.+", ".", acnt_base_name)
    acnt_file_path = os.path.join(output_dir, acnt_base_name)
    logger.info("Writing inferred diploid adjacency copy number data to {file}".format(file=acnt_file_path))
    write_acnt_to_file(file_name=acnt_file_path, acnt=acnt, adjacencies=adjacencies, output_reference=True, mix_reference_and_novel=args.o_acnt_mix_novel_and_reference)

    #########
    #
    # POST PROCESSING
    #
    #########

    if args.post_check_all or args.post_check_scnb:
        logger.info("Performing post-inference check that inferred segment copy number values are within input and/or preprocessed bounds")
        for clone_id in clone_ids:
            logger.info("Working with clone {clone_id}".format(clone_id=clone_id))
            scnp = scnt[clone_id]
            scnbp = scnb[clone_id]
            problems = False
            for segment in segments:
                sid = segment.stable_id_non_hap
                sync_indicator = ilp_model.alleles_sync_result(segment=segment)
                cna, cnb = scnp.get_cn(sid=sid, haplotype=Haplotype.A), scnp.get_cn(sid=sid, haplotype=Haplotype.B)
                lower_a = scnbp.get_cnb(sid=sid, hap=Haplotype.A, boundary_type=CNBoundaries.LOWER)
                lower_b = scnbp.get_cnb(sid=sid, hap=Haplotype.B, boundary_type=CNBoundaries.LOWER)
                upper_a = scnbp.get_cnb(sid=sid, hap=Haplotype.A, boundary_type=CNBoundaries.UPPER)
                upper_b = scnbp.get_cnb(sid=sid, hap=Haplotype.B, boundary_type=CNBoundaries.UPPER)
                if sync_indicator == 1:
                    if (not lower_a <= cna <= upper_a) or (not lower_b <= cnb <= upper_b):
                        problems = True
                else:
                    if (not lower_b <= cna <= upper_b) or (not lower_a <= cnb <= upper_a):
                        problems = True
                if problems:
                    logger.error("Something went WRONG! For segment {sid} with allele-sync flag {flag} inferred copy numbers are A={cna}, B={cnb},"
                                 "the input copy number boundaries were A={lower_a}-{upper_a}, B={lower_b}-{upper_b}."
                                 "".format(sid=sid, flag=sync_indicator, cna=cna, cnb=cnb, lower_a=lower_a, upper_a=upper_a, lower_b=lower_b, upper_b=upper_b))
            if not problems:
                logger.info("Everything is OK!")

    if args.post_check_all or args.post_check_labeling:
        logger.info("Performing post-inference check on labels on adjacencies with positive copy numbers")
        logger.info("Checking that every unlabeled novel adjacency has at most one labeled 'realization' across all clones")
        novel_adjacencies = [a for a in adjacencies if a.adjacency_type == AdjacencyType.NOVEL]
        violations = False
        for adj in novel_adjacencies:
            labeling_realizations = set()
            for clone_id in clone_ids:
                acnp = acnt[clone_id]
                for ph in [Phasing.AA, Phasing.AB, Phasing.BA, Phasing.BB]:
                    if acnp.get_cn(aid=adj.stable_id_non_phased, phasing=ph) > 0:
                        labeling_realizations.add(ph)
            if len(labeling_realizations) > 1:
                violations = True
                logger.error("Something went WRONG! For unlabeled novel adjacency (id = {aid}) {stable_id} more than 1 labeled 'realization' is present across all clones"
                             "".format(aid=adj.extra.get(EXTERNAL_NA_ID, adj.idx), stable_id=adj.stable_id_non_phased))
        if not violations:
            logger.info("Everything is OK! No violations of multiple 'realizations' were identified.")

        logger.info("Checking that every reference location, that has reciprocal novel adjacencies, concurs with the Infinite Sites constraints")
        combined_scnp = SegmentCopyNumberProfile.combined(*[scnt[clone_id] for clone_id in clone_ids])
        combined_acnp = AdjacencyCopyNumberProfile.combined(*[acnt[clone_id] for clone_id in clone_ids])
        hiag = construct_hiag_inflate_from_haploid_data(hapl_segments=segments, hapl_adjacencies=adjacencies)
        hiag.assign_copy_numbers_from_scn_profile(scn_profile=combined_scnp)
        hiag.assign_copy_numbers_from_acn_profile(acn_profile=combined_acnp)
        hiag.remove_edges_with_zero_cn()

        logger.info("Checking for extremity-exclusivity constraint violations")
        extremity_exclusivity_violations = hiag.violations_of_extremity_exclusivity()
        if len(extremity_exclusivity_violations) > 0:
            logger.error("Something went WRONG! The following segments extremities violate extremity-exclusivity constraint of the Infinite Sites model.")
            logger.error(",".join(map(str, extremity_exclusivity_violations)))
        else:
            logger.info("Everything is OK! No extremity-exclusivity violations were found!")
        logger.info("Checking for homologous-extremity-exclusivity violations")
        homologous_extremity_exclusivity_violations = hiag.violations_of_homologous_extremity_exclusivity()
        if len(homologous_extremity_exclusivity_violations) > 0:
            logger.error("Something went WRONG! The following pairs of segments extremities violate homologous-extremity-exclusivity constraint of the Infinite Sites model.")
            string_data = []
            for p1, p2 in homologous_extremity_exclusivity_violations:
                string_data.append("{{{p1},{p2}}}".format(p1=str(p1), p2=str(p2[0])))
            logger.error(" , ".join(string_data))
        else:
            logger.info("Everything is OK! No homologous-extremity-exclusivity violations were found!"
                        "".format(cnt=len(ilp_model.reciprocal_locations)))
        logger.info("Checking for reciprocal-homologous-extremity-exclusivity violations")
        homologous_reciprocal_extremity_exclusivity_violations = hiag.violations_of_homologous_reciprocal_extremity_exclusivity()
        if len(homologous_reciprocal_extremity_exclusivity_violations) > 0:
            logger.error("Something went WRONG! "
                         "The following pairs of segments extremities violate homologous-reciprocal-extremity-exclusivity constraint of the Infinite Sites model.")
            string_data = []
            for p1, p2 in homologous_reciprocal_extremity_exclusivity_violations:
                string_data.append("{{{p1},{p2}}}".format(p1=str(p1), p2=str(p2)))
            logger.error(" , ".join(string_data))
        else:
            logger.info("Everything is OK! Ho homologous-reciprocal-extremity-exclusivity were found across {cnt} reciprocal locations!"
                        "".format(cnt=len(ilp_model.reciprocal_locations)))

    if args.post_check_all or args.post_check_balancing:
        logger.info("Performing post-inference check on balances/excesses on segments' extremities")
        for clone_id in clone_ids:
            logger.info("Processing clone {clone_id}".format(clone_id=clone_id))
            hiag = construct_hiag_inflate_from_haploid_data(hapl_segments=segments, hapl_adjacencies=adjacencies)
            scnp = scnt[clone_id]
            acnp = acnt[clone_id]
            hiag.assign_copy_numbers_from_scn_profile(scn_profile=scnp)
            hiag.assign_copy_numbers_from_acn_profile(acn_profile=acnp)
            hiag.remove_edges_with_zero_cn()
            logger.info("Checking that every vertex has a copy number excess >= 0.")
            for node in hiag.nodes(data=False):
                if hiag.node_imbalance(node=node) < 0:
                    logger.error("Something went WRONG! On segment extremity {node} there is a negative copy number excess...")
                    exit(1)
            logger.info("Getting inferred telomeres.")
            diploid_telomeres = hiag.get_telomeres()
            inferred_hapl_telomeres_ids = {p.stable_id_non_hap for p in diploid_telomeres}
            input_hapl_telomers_ids = {p.stable_id_non_hap for p in telomeres}
            if inferred_hapl_telomeres_ids > input_hapl_telomers_ids:
                logger.error("Something went WRONG! Following segments extremities, while not specified specified as possible telomere sites were inferred as such.")
                logger.error(",".join(map(str, sorted(input_hapl_telomers_ids - input_hapl_telomers_ids))))
            else:
                logger.info("Everything is OK! in clone {clone_id} all extremities have non-negative copy number excess, and inferred telomere sites concur with the input"
                            "".format(clone_id=clone_id))

    if args.post_check_all or args.post_check_adj_groups or args.post_check_adj_groups_m:
        logger.info("Performing post-inference check on adjacencies groups (molecule type)")
        molecule_groups = [ag for ag in adjacency_groups if ag.group_type == AdjacencyGroupType.MOLECULE]
        molecule_groups_fine_cnt = 0
        logger.info("There were {cnt} molecule adjacency groups in the input.".format(cnt=len(molecule_groups)))
        for group in molecule_groups:
            group_fp = group.extra.get(FALSE_POSITIVE, extra[DEFAULT_GROUP_M_FP])
            group_is_good = False
            good_inferred_fp = -1
            for clone_id in clone_ids:
                acnp = acnt[clone_id]
                clone_present = acnp.haploid_adjacencies_present(adjacencies=group.adjacencies)
                inferred_fp = 1 - (len(clone_present) * 1.0 / len(group.adjacencies))
                if inferred_fp <= group_fp:
                    good_inferred_fp = inferred_fp
                group_is_good |= inferred_fp <= group_fp
            if not group_is_good:
                logger.error("Something went WRONG! In Adjacency Group {gid} not a single clone has <= than the input FP of {input_fp}."
                             "".format(gid=group.gid, input_fp=group_fp))
            else:
                logger.debug("Group {gid} is fine (there is a clone with group FP of {fp:0.4f} which is >= than the input FP of {ifp:0.4f})."
                             "".format(gid=group.gid, fp=good_inferred_fp, ifp=group.extra.get(FALSE_POSITIVE, extra[DEFAULT_GROUP_M_FP])))
                molecule_groups_fine_cnt += 1
        if molecule_groups_fine_cnt == len(molecule_groups):
            logger.info("Everything is OK for all {good}/{all} molecule adjacency groups.".format(good=molecule_groups_fine_cnt, all=len(molecule_groups)))
        else:
            logger.error("Something went WRONG! In some molecule adjacency groups (see above) not a single clone has <= than the input FP of.")

    if args.post_check_all or args.post_check_nas_fp:
        logger.info("Performing post-inference check on overall novel adjacencies false positive parameter")
        novel_adjacencies = [a for a in adjacencies if a.adjacency_type == AdjacencyType.NOVEL]
        present = set()
        for clone_id in clone_ids:
            acnp = acnt[clone_id]
            clone_present = acnp.haploid_adjacencies_present(adjacencies=novel_adjacencies)
            clone_present_ids = {adj.extra.get(EXTERNAL_NA_ID, adj.idx) for adj in clone_present}
            present |= clone_present_ids
        inferred_fp = 1 - (len(present) * 1.0 / len(novel_adjacencies))
        if inferred_fp <= overall_nas_fp:
            logger.info("Everything is OK! Inferred FP = {inf_fp:0.4f}, while the input FP was {input_fp:0.4f}.".format(inf_fp=inferred_fp, input_fp=overall_nas_fp))
        else:
            logger.error("Something went WRONG! Inferred FP = {inf_fp:0.4f}, while the input FP was {input_fp:0.4f}.".format(inf_fp=inferred_fp, input_fp=overall_nas_fp))
        logger.info("Finished everything. Hope you've enjoyed using RCK!")


if __name__ == "__main__":
    main()



import argparse
import sys
import os

current_file_level = 3
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

from rck.core.io import get_logging_cli_parser, get_standard_logger_from_args, write_scnt_to_destination, get_full_path
from rck.utils.scn.convert import get_scnt_from_battenberg_source, get_scnt_from_hatchet_source, hatchet_get_clone_ids_from_file, \
    get_scnt_from_remixt_source, titan_get_clone_ids_from_file, get_scnt_from_titan_source
from rck.utils.adj.process import get_chromosome_strip_parser


def main():
    parser = argparse.ArgumentParser(prog="RCK-UTILS-SCNT-x2rck")
    cli_logging_parser = get_logging_cli_parser()
    chr_strip_parser = get_chromosome_strip_parser()
    subparsers = parser.add_subparsers(title="command", dest="command")
    subparsers.required = True
    ####
    titan_parser = subparsers.add_parser("titan", parents=[cli_logging_parser, chr_strip_parser])
    titan_parser.add_argument("titan_ichor_seg")
    titan_parser.add_argument("--sample-name", required=True)
    titan_parser.add_argument("--clone-ids", default=None)
    titan_parser.add_argument("--separator", default="\t")
    titan_parser.add_argument("--corrected-cn-fix", choices=["None", "equal", "relative-dist"], default="None")
    titan_parser.add_argument("-o", "--output", type=argparse.FileType("wt"), default=sys.stdout)
    ####
    battenberg_parser = subparsers.add_parser("battenberg", parents=[cli_logging_parser, chr_strip_parser])
    battenberg_parser.add_argument("battenberg", type=argparse.FileType("rt"), default=sys.stdin)
    battenberg_parser.add_argument("--separator", default="\t")
    battenberg_parser.add_argument("--sample-name", required=True)
    battenberg_parser.add_argument("--clone-ids", choices=["1", "2", "1,2"], default="1,2")
    battenberg_parser.add_argument("-o", "--output", type=argparse.FileType("wt"), default=sys.stdout)
    ####
    hatchet_parser = subparsers.add_parser("hatchet", parents=[cli_logging_parser, chr_strip_parser])
    hatchet_parser.add_argument("hatchet", type=str)
    hatchet_parser.add_argument("--separator", default="\t")
    hatchet_parser.add_argument("--min-usage", type=float, default=0.01)
    hatchet_parser.add_argument("-o", "--output", type=argparse.FileType("wt"), default=sys.stdout)
    group = hatchet_parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--sample-name", default=None)
    group.add_argument("--clone-ids", default=None)
    ####
    remixt_parser = subparsers.add_parser("remixt", parents=[cli_logging_parser, chr_strip_parser])
    remixt_parser.add_argument("remixt", type=argparse.FileType("rt"), default=sys.stdin)
    remixt_parser.add_argument("--separator", default="\t")
    remixt_parser.add_argument("--clone-ids", choices=["1", "2", "1,2"], default="1,2")
    remixt_parser.add_argument("-o", "--output", type=argparse.FileType("wt"), default=sys.stdout)
    ####
    args = parser.parse_args()
    logger = get_standard_logger_from_args(args=args, program_name="RCK-UTILS-SCNT")

    if args.command == "titan":
        logger.info("Converting allele-specific segment copy values form TitanCNA format to RCK")
        titan_full_path = get_full_path(path=args.titan_ichor_seg)
        if args.clone_ids is None:
            logger.debug("Clone ids were not provided, extracting all clone ids from {file}".format(file=titan_full_path))
            clone_ids = titan_get_clone_ids_from_file(file_name=titan_full_path, sample_name=args.sample_name, separator=args.separator)
        else:
            clone_ids = sorted(set(args.clone_ids.split(",")))
        logger.debug("Clone ids are identified as {clone_ids}".format(clone_ids=",".join(clone_ids)))
        with open(args.titan_ichor_seg, "rt") as source:
            logger.info("Reading allele-specific segment copy number values from {file}".format(file=titan_full_path))
            segments, scnt = get_scnt_from_titan_source(source=source, sample_name=args.sample_name, clone_ids=clone_ids, separator=args.separator,
                                                        corrected_cn_fix=args.corrected_cn_fix, chr_strip=args.strip_chr)
            logger.info("Writing allele-specific segment copy number values in RCK format to {file}".format(file=args.output))
            write_scnt_to_destination(destination=args.output, segments=segments, scnt=scnt, clone_ids=clone_ids, separator=args.separator)
    elif args.command == "battenberg":
        logger.info("Converting allele-specific segment copy values form Battenberg format to RCK")
        clone_ids = args.clone_ids.split(",")
        logger.debug("Clone ids are identified as {clone_ids}".format(clone_ids=",".join(clone_ids)))
        logger.info("Reading allele-specific segment copy number values form {file}".format(file=args.battenberg))
        segments, scnt = get_scnt_from_battenberg_source(source=args.battenberg, sample_name=args.sample_name, separator=args.separator, chr_strip=args.strip_chr)
        logger.info("Writing allele-specific segment copy number values in RCK format to {file}".format(file=args.output))
        write_scnt_to_destination(destination=args.output, segments=segments, scnt=scnt, separator=args.separator, clone_ids=clone_ids)
    elif args.command == "hatchet":
        hatchet_full_path = get_full_path(path=args.hatchet)
        logger.info("Converting allele-specific segment copy values form HATCHet format to RCK")
        if args.clone_ids is None:
            logger.debug("Clone ids were not provided, extracting all clone ids from {file}".format(file=hatchet_parser))
            clone_ids = hatchet_get_clone_ids_from_file(file_name=hatchet_full_path, sample_name=args.sample_name, separator=args.separator, min_usage=args.min_usage)
        else:
            clone_ids = sorted(set(args.clone_ids.split(",")))
        logger.debug("Clone ids were identified as {clone_ids}".format(clone_ids=",".join(clone_ids)))
        with open(hatchet_full_path) as source:
            logger.info("Reading allele-specific segment copy number values from {file}".format(file=hatchet_full_path))
            segments, scnt = get_scnt_from_hatchet_source(source=source, clone_ids=clone_ids, separator=args.separator, chr_strip=args.strip_chr)
            logger.info("Writing allele-specific segment copy number values in RCK format to {file}".format(file=args.output))
            write_scnt_to_destination(destination=args.output, segments=segments, scnt=scnt, clone_ids=clone_ids, separator=args.separator)
    elif args.command == "remixt":
        logger.info("Converting allele-specific segment copy values form ReMixT format to RCK")
        clone_ids = args.clone_ids.split(",")
        logger.debug("Clone ids were identified as {clone_ids}".format(clone_ids=",".join(clone_ids)))
        logger.info("Reading allele-specific segment copy number values from {file}".format(file=args.remixt))
        segments, scnt = get_scnt_from_remixt_source(source=args.remixt, separator=args.separator, chr_strip=args.strip_chr)
        logger.info("Writing allele-specific segment copy number values in RCK format to {file}".format(file=args.output))
        write_scnt_to_destination(destination=args.output, segments=segments, scnt=scnt, separator=args.separator, clone_ids=clone_ids)
    logger.info("Success!")


if __name__ == "__main__":
    main()



from setuptools import setup

import sys
import os

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import rck

setup(
    name="RCK",
    version=rck.version,
    author="Sergey Aganezov",
    author_email="aganezov@cs.jhu.edu",
    description="A tool for (R)econstruction of (C)ancer (K)aryotypes (both clone- and haplotype-specific)",
    license="MIT",
    keywords="RCK, rck, cancer, cancer genomics, cancer karyotypes, clonality, subclonality, copy number aberrations, breakpoints, structural variations, novel adjacencies",
    url="",
    packages=["", "rck", "rck.core", "rck.utils", "rck.utils.scn", "rck.utils.adj"],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "rck = rck.rck_run:main",
            "rck-scnt-x2rck = rck.utils.scn.rck_scnt_x2rck:main",
            "rck-scnt-process = rck.utils.scn.rck_scnt_process:main",
            "rck-scnb = rck.utils.scn.rck_scnb:main",
            "rck-adj-x2rck = rck.utils.adj.rck_adj_x2rck:main",
            "rck-adj-rck2x = rck.utils.adj.rck_adj_rck2x:main",
            "rck-adj-process = rck.utils.adj.rck_adj_process:main",
            "rck-input-refine = rck.utils.rck_input_refine:main",
        ]
    },
    install_requires=[
        "networkx>=2",
        "scipy",
        "pyvcf",
        "pysam",
        "sortedcontainers",
        "pandas",
    ]
)


import logging
from collections import defaultdict
import networkx as nx
import pysam

from rck.core.io import read_adjacencies_from_source

#####


def get_mode_str(format="bam", input=False):
    result = "r" if input else "w"
    if format == "bam":
        result += "b"
    elif format == "cram":
        result += "c"
    return result

#####


def extract_long_reads():
    pass


def filter_alignment():
    pass


def infer_labeling_constraints(rck_nas_source, alignment_file, i_alignment_format, lr_field, min_sv_cnt, logger=None):
    logger = logger or logging.getLogger('dummy')
    logger.info("RCK NAs file object {file_name}".format(file_name=str(rck_nas_source)))
    nas = read_adjacencies_from_source(source=rck_nas_source)
    reads_to_nas = defaultdict(list)
    for na in nas:
        reads_str = na.extra.get(lr_field, "")
        reads = reads_str.split(",")
        for read in reads:
            if len(read) == 0:
                continue
            reads_to_nas[read].append(na)
    logger.debug("{reads_cnt} -- number of reads".format(reads_cnt=len(reads_to_nas)))
    reads = {read for read in reads_to_nas if len(reads_to_nas[read]) >= min_sv_cnt}
    logger.debug("{reads_cnt} -- number of reads that each span {min_sv_cnt}+ NAs".format(reads_cnt=len(reads), min_sv_cnt=min_sv_cnt))
    location_graph = nx.Graph()
    mode = get_mode_str(format=i_alignment_format, input=True)
    current_read_name = None
    current_entries = []
    with pysam.AlignmentFile(alignment_file, mode) as i_stream:
        if "SO:queryname" not in i_stream.text:
            logger.critical("Input alignment file {alignment_file} is not sorted by read (i.e., query) name".format(alignment_file=alignment_file))
            raise Exception("Input bam file needs to be sorted by read (i.e., query) name")
        for entry in i_stream:
            if entry.qname != current_read_name:
                if len(current_entries) > 0:
                    reads_novel_adjacencies = reads_to_nas[current_read_name]
                    add_lr_labeling_constraints(location_graph=location_graph, alignment_entries=current_entries, nas=reads_novel_adjacencies)
                current_read_name = entry.qname
                current_entries = [entry]
            else:
                current_entries.append(entry)
        # last reads streak has to be processed as well
        if current_read_name in reads_to_nas:
            pass


def add_lr_labeling_constraints(location_graph, alignment_entries, nas):
    entries = sorted(alignment_entries, key=lambda e: (e.query_alignment_start, e.query_alignment_end))
    internal = len(nas) != len(entries) - 1

    pass


def label_constraints_combining():
    pass




import argparse

import sys
import os

current_file_level = 3
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

import rck
from rck.core.io import get_logging_cli_parser, get_standard_logger_from_args
from rck.utils.adj.convert import *
from rck.utils.adj.process import filter_adjacencies_by_chromosomal_regions, get_shared_nas_parser, processed_gundem2015_adjacencies, get_chromosome_strip_parser


def main():
    parser = argparse.ArgumentParser(prog="RCK-UTILS-ADJ-x2rck")
    parser.add_argument('--version', action='version', version=rck.version)
    ####
    shared_parser = get_shared_nas_parser()
    shared_parser.add_argument("--output", "-o", dest="rck_adj_file", type=argparse.FileType("wt"), default=sys.stdout)
    cli_logging_parser = get_logging_cli_parser()
    chr_strip_parser = get_chromosome_strip_parser()
    ####
    subparsers = parser.add_subparsers(title="commands", dest="command")
    subparsers.required = True
    ####
    lumpy_parser = subparsers.add_parser("lumpy", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert Lumpy VCF SV calls into RCK NAS format")
    lumpy_parser.add_argument("--id-suffix", dest="id_suffix", default="lumpy")
    lumpy_parser.add_argument("lumpy_vcf_file", type=argparse.FileType("rt"), default=sys.stdin)
    ####
    longranger_parser = subparsers.add_parser("longranger", parents=[shared_parser, cli_logging_parser, chr_strip_parser],
                                              help="Convert LongRanger VCF SV calls into RCK NAS format")
    longranger_parser.add_argument("--id-suffix", dest="id_suffix", default="longranger")
    longranger_parser.add_argument("longranger_vcf_file", type=argparse.FileType("rt"), default=sys.stdin)
    ####
    naibr_parser = subparsers.add_parser("naibr", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert NAIBR NAS calls into RCK NAS format")
    naibr_parser.add_argument("--id-suffix", dest="id_suffix", default="naibr")
    naibr_parser.add_argument("naibr_file", type=argparse.FileType("rt"), default=sys.stdin)
    ####
    manta_parser = subparsers.add_parser("manta", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert Manta VCF SV calls into RCK NAS format")
    manta_parser.add_argument("--id-suffix", dest="id_suffix", default="manta")
    manta_parser.add_argument("manta_vcf_file", type=argparse.FileType("rt"), default=sys.stdin)
    ####
    sniffles_parser = subparsers.add_parser("sniffles", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert Sniffles VCF SV calls into RCK NAS format")
    sniffles_parser.add_argument("--id-suffix", dest="id_suffix", default="sniffles")
    sniffles_parser.add_argument("sniffles_vcf_file", type=argparse.FileType("rt"), default=sys.stdin)
    ####
    grocsv = subparsers.add_parser("grocsv", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert GROCSVS VCF SV calls into RCK NAS format")
    grocsv.add_argument("--id-suffix", dest="id_suffix", default="grocsv")
    grocsv.add_argument("grocsv_vcf_file", type=argparse.FileType("rt"), default=sys.stdin)
    grocsv.add_argument("--samples")
    grocsv.add_argument("--samples-all-any", choices=["all", "any"], default="any")
    grocsv.add_argument("--samples-only", action="store_true", dest="samples_only")
    ####
    delly = subparsers.add_parser("delly", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert Delly VCF SV calls into RCK NAS format")
    delly.add_argument("--id-suffix", dest="id_suffix", default="delly")
    delly.add_argument("delly_vcf_file", type=argparse.FileType("rt"), default=sys.stdin)
    delly.add_argument("--stream", action="store_true", dest="delly_force_stream")
    ####
    pbsv = subparsers.add_parser("pbsv", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert PBSV VCF SV calls into RCK NAS format")
    pbsv.add_argument("--id-suffix", dest="id_suffix", default="pbsv")
    pbsv.add_argument("pbsv_vcf_file", type=argparse.FileType("rt"), default=sys.stdin)
    ####
    remixt = subparsers.add_parser("remixt", parents=[shared_parser, cli_logging_parser, chr_strip_parser], help="Convert ReMixT Novel adjacencies calls into RCK format")
    remixt.add_argument("--i-separator", default="\t")
    remixt.add_argument("--id-suffix", dest="id_suffix", default="remixt")
    remixt.add_argument("--clone-ids", choices=["1", "2", "1,2"], default="1,2")
    remixt.add_argument("--skip-absent", action="store_true", dest="skip_absent")
    remixt.add_argument("--no-remixt-na-correction", action="store_false", dest="remixt_correction")
    remixt.add_argument("remixt_file", type=argparse.FileType("rt"), default=sys.stdin)
    ####
    gundem2015_parser = subparsers.add_parser("gundem2015", parents=[shared_parser, cli_logging_parser, chr_strip_parser],
                                              help="Convert SV calls from Gundem et al (2015) (BRASS2???) "
                                                   "into RCK NAS format")
    gundem2015_parser.add_argument("--id-suffix", dest="id_suffix", default="gundem2015")
    gundem2015_parser.add_argument("gundem2015_file", type=argparse.FileType("rt"), default=sys.stdin)
    gundem2015_parser.add_argument("--i-separator", default="\t")
    gundem2015_parser.add_argument("--samples", nargs="+", required=True)
    gundem2015_parser.add_argument("--min-sample-cnt", type=int, default=1)
    gundem2015_parser.add_argument("--no-flip-second-strand", action="store_false", dest="flip_second_strand")
    ####
    args = parser.parse_args()
    setup = build_setup(args=args)
    logger = get_standard_logger_from_args(args=args, program_name="RCK-UTILS-ADJ-x2rck")
    nas = []
    if args.o_extra_fields is None or len(args.o_extra_fields) == 0 or args.o_extra_fields == ",":
        extra = None
    elif args.o_extra_fields != "all":
        extra = args.o_extra_fields.split(",")
    else:
        extra = args.o_extra_fields
    if args.command == "lumpy":
        logger.info("Starting converting adjacencies from the Lumpy VCF format to that of RCK")
        logger.info("Reading Lumpy VCF records from {file}".format(file=args.lumpy_vcf_file))
        lumpy_vcf_records = get_vcf_records_from_source(source=args.lumpy_vcf_file)
        logger.info("Converting Lumpy VCF records to RCK adjacencies")
        nas = get_nas_from_lumpy_vcf_records(lumpy_vcf_records=lumpy_vcf_records, setup=setup)
    elif args.command == "longranger":
        logger.info("Starting converting adjacencies from the LongRanger VCf format to that of RCK")
        logger.info("Reading LongRanger VCF records from {file}".format(file=args.longranger_vcf_file))
        longranger_vcf_records = get_vcf_records_from_source(source=args.longranger_vcf_file)
        logger.info('Converting LongRanger VCF records to RCK adjacencies')
        nas = get_nas_from_longranger_vcf_records(longranger_vcf_records=longranger_vcf_records, setup=setup)
    elif args.command == "naibr":
        logger.info("Starting converting adjacencies from NAIBR records to that of RCK")
        logger.info("Reading and converting NAIBR records from {file} to RCK adjacencies".format(file=args.naibr_file))
        nas = get_nas_from_naibr_source(source=args.naibr_file, setup=setup)
    elif args.command == "manta":
        logger.info("Starting converting adjacencies from Manta records to that of RCK")
        logger.info("Reading Manta VCF records from {file}".format(file=args.manta_vcf_file))
        manta_vcf_records = get_vcf_records_from_source(source=args.manta_vcf_file)
        logger.info("Converting Manta VCF records to RCK adjacencies")
        nas = get_nas_from_manta_vcf_records(manta_vcf_records=manta_vcf_records, setup=setup)
    elif args.command == "sniffles":
        logger.info("Starting converting adjacencies from Sinffles records to that of RCK")
        logger.info("Reading Sniffles VCF records from {file}".format(file=args.sniffles_vcf_file))
        sniffles_vcf_records = get_vcf_records_from_source(source=args.sniffles_vcf_file)
        logger.info("Converting Sniffles VCF records to RCK adjacencies")
        nas = get_nas_from_sniffles_vcf_records(sniffles_vcf_records=sniffles_vcf_records, setup=setup)
    elif args.command == "grocsv":
        logger.info("Starting converting adjacencies from GROCSVS records to that of RCK")
        logger.info("Reading GROCSVS VCF records from {file}".format(file=args.grocsv_vcf_file))
        samples = args.samples.split(",") if args.samples is not None else args.samples
        grocsv_vcf_records = get_vcf_records_from_source(source=args.grocsv_vcf_file)
        logger.info("Converting GROCSVS VCF records to RCK adjacencies")
        nas = get_nas_from_grocsv_vcf_records(grocsv_vcf_records=grocsv_vcf_records, setup=setup, samples=samples, sample_all_any=args.samples_all_any,
                                              samples_only=args.samples_only)
    elif args.command == "delly":
        logger.info("Starting converting adjacencies from Delly records to that of RCK")
        if args.delly_force_stream:
            logger.info("Forced stream is enabled")
            logger.info("Streamlining reading and converting Delly VCF from {i_file} to RCK adjacencies into {o_file}"
                        "".format(i_file=args.delly_vcf_file, o_file=args.rck_nas_file))
            delly_vcf_to_nas_stream(source=args.delly_vcf_file, dest=args.rck_nas_file, setup=setup, extra=extra)
            sys.exit(0)
        else:
            logger.info("Reading Delly VCF records from {file}".format(file=args.delly_vcf_file))
            delly_vcf_records = get_vcf_records_from_source(source=args.delly_vcf_file)
            logger.info("Converting Delly VCF records to rCK adjacencies")
            nas = get_nas_from_delly_vcf_records(delly_vcf_records=delly_vcf_records, setup=setup)
    elif args.command == "pbsv":
        logger.info("Starting converting adjacencies from PBSV records to that of RCK")
        logger.info("Reading PBSV VCF records from {file}".format(file=args.pbsv_vcf_file))
        pbsv_vcf_records = get_vcf_records_from_source(source=args.pbsv_vcf_file)
        logger.info("Converting PBSV VCF records to RCK adjacencies")
        nas = get_nas_from_pbsv_vcf_records(pbsv_vcf_records=pbsv_vcf_records, setup=setup)
    elif args.command == "gundem2015":
        logger.info("Starting converting adjacencies from Gundem et al 2015 (BRASS2???) to that of RCK")
        logger.info("Reading Gundem 2015 et al (BRASS???) records from {file}".format(file=args.gundem2015_file))
        nas = get_nas_from_gundem2015_source(source=args.gundem2015_file, setup=setup, separator=args.i_separator, flip_second_strand=args.flip_second_strand)
        logger.info("Extracting adjacencies for sample {samples} with a minimum cnt of {min_cnt}".format(samples=",".join(args.samples), min_cnt=args.min_sample_cnt))
        nas = processed_gundem2015_adjacencies(adjacencies=nas, sample_names=args.samples, min_per_sample_cnt=args.min_sample_cnt)
    elif args.command == "remixt":
        logger.info("Starting converting adjacencies and their (haploid) copy numbers from ReMixT to that of RCK")
        logger.info("Reading and converting ReMixT resotds from {file} to RCK adjacencies".format(file=args.remixt_file))
        clone_ids = args.clone_ids.split(",")
        nas = get_nas_from_remixt_source(source=args.remixt_file, setup=setup, separator=args.i_separator, clone_ids=clone_ids, skip_absent=args.skip_absent,
                                         remixt_na_correction=args.remixt_correction)
    logger.info("A total of {cnt} adjacencies were obtained.".format(cnt=len(nas)))
    logger.debug("Output extra fields were identified as {o_extra}".format(o_extra=",".join(extra)))
    include_chrs_regions_strings = []
    exclude_chrs_regions_strings = []
    if args.chrs_include is not None:
        for chrs_lists in args.chrs_include:
            for chrs_list in chrs_lists:
                for chr_name in chrs_list.split(","):
                    include_chrs_regions_strings.append(chr_name)
    if args.chrs_include_file is not None:
        for chr_name in get_chrs_regions_string_lists_from_source(source=args.chrs_include_file):
            include_chrs_regions_strings.append(chr_name)
    if args.chrs_exclude is not None:
        for chrs_lists in args.chrs_exclude:
            for chrs_list in chrs_lists:
                for chr_name in chrs_list.split(","):
                    exclude_chrs_regions_strings.append(chr_name)
    if args.chrs_exclude_file is not None:
        for chr_name in get_chrs_regions_string_list_from_file(file_name=args.chrs_exclude_file):
            exclude_chrs_regions_strings.append(chr_name)
    include_regions = [parse_segment_chr_region(string) for string in include_chrs_regions_strings]
    exclude_regions = [parse_segment_chr_region(string) for string in exclude_chrs_regions_strings]
    logger.debug("Include chromosomes : {include_chromosomes}".format(include_chromosomes=",".join(map(str, include_regions))))
    logger.debug("Exclude chromosomes : {exclude_chromosomes}".format(exclude_chromosomes=",".join(map(str, exclude_regions))))
    logger.info("Filtering adjacencies based on input/exclude chromosomes")
    nas = filter_adjacencies_by_chromosomal_regions(adjacencies=nas, include=include_regions, exclude=exclude_regions, include_both=args.include_both, exclude_both=args.exclude_both)
    nas = list(nas)
    logger.info("A total of {cnt} adjacencies were retained after filtering".format(cnt=len(nas)))
    logger.info("Writing RCK adjacencies to {file}".format(file=args.rck_adj_file))
    write_adjacencies_to_destination(destination=args.rck_adj_file, adjacencies=nas, extra=extra)
    logger.info("Success")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import argparse
import ast
import csv
import sys
import datetime
import itertools
import logging
import os
from collections import defaultdict
from copy import deepcopy
from enum import Enum

from rck.core.structures import AdjacencyCopyNumberProfile, AdjacencyGroup, CNBoundaries, SegmentCopyNumberBoundaries, AdjacencyGroupType
from rck.core.structures import SegmentCopyNumberProfile, Haplotype, AdjacencyType, Phasing
from rck.core.structures import Position, Strand, Adjacency, Segment

csv.field_size_limit(sys.maxsize)

AID = "aid"
GID = "gid"
AIDS = "aids"
AG_TYPE = "agt"
AG_LABELING = "agl"
FALSE_POSITIVE = "fp"

CHR1 = "chr1"
COORD1 = "coord1"
CHR2 = "chr2"
COORD2 = "coord2"
STRAND1 = "strand1"
STRAND2 = "strand2"
EXTRA = "extra"
COPY_NUMBER = "cn"
OLD_SEGMENTS_CNS = "CNs"
OLD_ADJACENCIES_CNS = "CNs"
COPY_NUMBER_BOUNDARIES = "cnb"
ADJACENCY_TYPE = "at"

CHR = "chr"
START = "start"
END = "end"
COORD = "coord"
STRAND = "strand"

VCF_GENOTYPE = "GT"


def get_logging_cli_parser():
    logging_parser = argparse.ArgumentParser(add_help=False)
    logging_parser.add_argument("--log-level", type=int,
                                choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL], default=logging.INFO)
    logging_parser.add_argument("--log-format", default="%(asctime)s - %(name)-15s - %(levelname)-7s - %(message)s")
    logging_parser.add_argument("--log-file", default=None)
    return logging_parser


def get_standard_logger_from_args(args, program_name):
    logger = logging.getLogger(program_name)
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(args.log_format)
    ch = logging.StreamHandler()
    ch.setLevel(args.log_level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    if args.log_file is not None:
        full_path = get_full_path(path=args.log_file)
        fh = logging.FileHandler(full_path)
        fh.setLevel(args.log_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    return logger


def extra_description_is_all(extra):
    return (len(extra) == 1 and extra[0].lower() == "all") or (isinstance(extra, str) and extra.lower() == "all")


def parse_cn_entry(cn_string, cn_separator=";"):
    result = dict()
    clone_specific_entries = cn_string.split(cn_separator)
    for entry in clone_specific_entries:
        entry = entry[1:-1]
        clone_id, cna, cnb = entry.split(",")
        cna, cnb = int(cna), int(cnb)
        result[clone_id] = {Haplotype.A: cna, Haplotype.B: cnb}
    return result


def parse_scn_record(scn_string, separator="\t", cn_separator=";", allow_unit_segments=False):
    data = scn_string.split(separator)
    chr_name = data[0]
    start_coordinate = int(data[1])
    end_coordinate = int(data[2])
    cn_entry = data[3]
    start_position = Position(chromosome=chr_name, coordinate=start_coordinate, strand=Strand.REVERSE)
    end_position = Position(chromosome=chr_name, coordinate=end_coordinate, strand=Strand.FORWARD)
    segment = Segment(start_position=start_position, end_position=end_position, allow_unit_length=allow_unit_segments)
    cns = parse_cn_entry(cn_string=cn_entry, cn_separator=cn_separator)
    return segment, cns


def read_scn_tensor(file_name, clone_ids, separator="\t", cn_separator=";", allow_absent_clones=False, allow_unit_segments=False):
    segments = []
    result = {}
    for clone_id in clone_ids:
        result[clone_id] = SegmentCopyNumberProfile()
    with open(file_name, "rt") as source:
        for line_cnt, line in enumerate(source):
            line = line.strip()
            if len(line) == 0 or line.startswith("#") or line_cnt == 0:
                continue
            segment, cns = parse_scn_record(scn_string=line, separator=separator, cn_separator=cn_separator, allow_unit_segments=allow_unit_segments)
            for clone_id in clone_ids:
                if not allow_absent_clones and clone_id not in cns:
                    raise Exception()
            segments.append(segment)
            sid = segment.stable_id_non_hap
            for clone_id in clone_ids:
                scnp = result[clone_id]
                cna = cns[clone_id][Haplotype.A]
                cnb = cns[clone_id][Haplotype.B]
                scnp.set_cn_record(sid=sid, hap=Haplotype.A, cn=cna)
                scnp.set_cn_record(sid=sid, hap=Haplotype.B, cn=cnb)
    return segments, result


def get_cn_string_entry(clone_id, cna, cnb):
    return "({clone_id},{cna},{cnb})".format(clone_id=str(clone_id), cna=str(cna), cnb=str(cnb))


def get_scn_string_entry(segment, scnt, clone_ids, separator="\t", cn_separator=";"):
    chr_string = "{chr_name}".format(chr_name=str(segment.chromosome))
    start_coord_string = "{start_coord}".format(start_coord=str(segment.start_position.coordinate))
    end_coord_string = "{end_coord}".format(end_coord=str(segment.end_position.coordinate))
    cn_strings = []
    sid = segment.stable_id_non_hap
    for clone_id in clone_ids:
        scnp = scnt[clone_id]
        cna = scnp.get_cn(sid=sid, haplotype=Haplotype.A, default=0)
        cnb = scnp.get_cn(sid=sid, haplotype=Haplotype.B, default=0)
        cn_string = get_cn_string_entry(clone_id=clone_id, cna=cna, cnb=cnb)
        cn_strings.append(cn_string)
    cn_string = cn_separator.join(cn_strings)
    all_string_entries = [chr_string, start_coord_string, end_coord_string, cn_string]
    return separator.join(all_string_entries)


def write_scn_tensor(file_name, scnt, segments, separator="\t", cn_separator=";"):
    clone_ids = sorted(scnt.keys())
    with open(file_name, "wt") as destination:
        print("chr", "start", "end", "CNs", sep=separator, file=destination)
        segments = sorted(segments, key=lambda s: (s.chromosome, s.start_position.coordinate, s.end_position.coordinate))
        for segment in segments:
            scn_string = get_scn_string_entry(segment=segment, scnt=scnt, clone_ids=clone_ids, separator=separator, cn_separator=cn_separator)
            print(scn_string, file=destination)


def read_scnt_from_file(file_name, clone_ids=None, separator="\t", extra_separator=";", remove_cn_data_from_segs=True):
    with open(file_name, "rt") as source:
        return read_scnt_from_source(source=source, clone_ids=clone_ids, separator=separator, extra_separator=extra_separator,
                                     remove_cn_data_from_segs=remove_cn_data_from_segs)


def read_scnt_from_source(source, clone_ids=None, separator="\t", extra_separator=";", remove_cn_data_from_segs=True):
    segments = read_segments_from_source(source=source, separator=separator, extra_separator=extra_separator)
    for segment in segments:
        if COPY_NUMBER not in segment.extra:
            raise ValueError("Trying to read Segment Copy Number tensor from {source}, but the {cn} entry in the {extra} column (or {extra} column is missing al together)"
                             "".format(source=source, cn=COPY_NUMBER, extra=EXTRA))
    scnt = extract_scnt_from_segments(segments=segments, clone_ids=clone_ids)
    if remove_cn_data_from_segs:
        remove_cn_data_from_segments(segments=segments)
    return segments, scnt


def remove_cn_data_from_segments(segments):
    for segment in segments:
        if COPY_NUMBER in segment.extra:
            del segment.extra[COPY_NUMBER]


def read_scnb_from_file(file_name, clone_ids=None, separator="\t", extra_separator=";", remove_cnb_data_from_segs=True, allow_missing=True):
    with open(file_name, "rt") as source:
        return read_scnb_from_source(source=source, clone_ids=clone_ids, separator=separator, extra_separator=extra_separator,
                                     remove_cnb_data_from_segs=remove_cnb_data_from_segs, allow_missing=allow_missing)


def read_scnb_from_source(source, clone_ids=None, separator="\t", extra_separator=";", remove_cnb_data_from_segs=True, allow_missing=True):
    segments = read_segments_from_source(source=source, separator=separator, extra_separator=extra_separator)
    if not allow_missing:
        for segment in segments:
            if COPY_NUMBER_BOUNDARIES not in segment:
                raise ValueError("Trying to read Segment Copy Number tensor from {source}, but the {cn} entry in the {extra} column (or {extra} column is missing al together)"
                                 "".format(source=source, cn=COPY_NUMBER, extra=EXTRA))
    scnb = extract_scnb_from_segments(segments=segments, clone_ids=clone_ids, allow_missing=allow_missing)
    if remove_cnb_data_from_segs:
        remove_cnb_data_from_segments(segments=segments)
    return segments, scnb


def remove_cnb_data_from_segments(segments):
    for segment in segments:
        if COPY_NUMBER_BOUNDARIES in segment.extra:
            del segment.extra[COPY_NUMBER_BOUNDARIES]


def extract_scnt_from_segments(segments, clone_ids=None):
    if clone_ids is None:
        clone_ids = set()
        for segment in segments:
            clone_ids.update(set(segment.extra[COPY_NUMBER].keys()))
    result = {clone_id: SegmentCopyNumberProfile() for clone_id in clone_ids}
    for segment in segments:
        cn_data = segment.extra[COPY_NUMBER]
        sid = segment.stable_id_non_hap
        for clone_id in clone_ids:
            if clone_id not in cn_data:
                raise ValueError("Clone {cid} is not present in CN data for segment {segment}".format(cid=clone_id, segment=sid))
            for haplotype in cn_data[clone_id]:
                scnp = result[clone_id]
                scnp.set_cn_record(sid=sid, hap=haplotype, cn=cn_data[clone_id][haplotype])
    return result


def extract_scnb_from_segments(segments, clone_ids=None, allow_missing=True):
    if clone_ids is None:
        clone_ids = set()
        for segment in segments:
            clone_ids.update(set(segment.extra[COPY_NUMBER].keys()))
    result = {clone_id: SegmentCopyNumberBoundaries() for clone_id in clone_ids}
    for segment in segments:
        if allow_missing and COPY_NUMBER_BOUNDARIES not in segment.extra:
            continue
        cnb_data = segment.extra[COPY_NUMBER_BOUNDARIES]
        sid = segment.stable_id_non_hap
        for clone_id in clone_ids:
            if clone_id not in cnb_data:
                raise ValueError("Clone {cid} is not present in CNB data for segment {segment}".format(cid=clone_id, segment=sid))
            for haplotype in cnb_data[clone_id]:
                for boundary_type in cnb_data[clone_id][haplotype]:
                    scnb = result[clone_id]
                    scnb.set_cnb_record(sid=sid, hap=haplotype, boundary_type=boundary_type, value=cnb_data[clone_id][haplotype][boundary_type])
    return result


def read_segments_from_file(file_name, separator="\t", extra_separator=";"):
    with open(file_name, "rt") as source:
        return read_segments_from_source(source=source, separator=separator, extra_separator=extra_separator)


def read_segments_from_source(source, separator="\t", extra_separator=";"):
    return list(stream_segments_from_source(source=source, separator=separator, extra_separator=extra_separator))


def parse_segment_extra_cn_boundaries_string(segment_cn_boundaries_string):
    clone_specific_dict = ast.literal_eval(node_or_string=segment_cn_boundaries_string)
    result = defaultdict(lambda: defaultdict(dict))
    for clone_id, data in clone_specific_dict.items():
        for haplotype_str in data:
            boundaries_dict = data[haplotype_str]
            haplotype = Haplotype[haplotype_str]
            for boundaries_string in boundaries_dict:
                boundary_value = int(boundaries_dict[boundaries_string])
                boundary = CNBoundaries.from_string(string=boundaries_string)
                result[clone_id][haplotype][boundary] = boundary_value
    return dict(result)


def parse_segment_extra_cn_string(segment_cn_string):
    clone_specific_dict = ast.literal_eval(node_or_string=segment_cn_string)
    result = defaultdict(dict)
    for clone_id, data in clone_specific_dict.items():
        for haplotype_str in data:
            cn = int(data[haplotype_str])
            haplotype = Haplotype[haplotype_str]
            result[clone_id][haplotype] = cn
    return dict(result)


def parse_segment_old_cn_string(old_cn_string):
    per_clone_entries = old_cn_string.split(";")
    result = {}
    for per_clone_entry in per_clone_entries:
        per_clone_entry = per_clone_entry[1:-1]
        clone, cna, cnb = per_clone_entry.split(",")
        cna, cnb = int(cna), int(cnb)
        result[clone] = {Haplotype.A: cna, Haplotype.B: cnb}
    return result


def parse_segment_extra(extra, extra_separator=";"):
    result = {}
    extra_entries = extra.split(extra_separator)
    for entry in extra_entries:
        data = entry.split("=")
        if len(data) == 1:
            if len(data[0]) > 0:
                result[data[0]] = True
            continue
        key, value = data
        if key == COPY_NUMBER:
            value = parse_segment_extra_cn_string(segment_cn_string=value)
        elif key == COPY_NUMBER_BOUNDARIES:
            value = parse_segment_extra_cn_boundaries_string(segment_cn_boundaries_string=value)
        result[key] = value
    return result


def stream_segments_from_source(source, separator="\t", extra_separator=";"):
    reader = csv.DictReader(source, delimiter=separator)
    for row in reader:
        chromosome = row[CHR]
        start = int(row[START])
        end = int(row[END])
        extra_dict = {}
        if OLD_SEGMENTS_CNS in row:
            extra_dict[COPY_NUMBER] = parse_segment_old_cn_string(old_cn_string=row[OLD_SEGMENTS_CNS])
        if EXTRA in row:
            supp_extra_dict = parse_segment_extra(extra=row[EXTRA], extra_separator=extra_separator)
            extra_dict.update(supp_extra_dict)
        segment = Segment.from_chromosome_coordinates(chromosome=chromosome, start=start, end=end)
        segment.extra = extra_dict
        yield segment


def write_segments_to_file(file_name, segments, separator="\t", extra="all", extra_separator=";", extra_fill="", sort_segments=True):
    with open(file_name, "wt") as destination:
        write_segments_to_destination(destination=destination, segments=segments, separator=separator,
                                      extra=extra, extra_separator=extra_separator, extra_fill=extra_fill, sort_segments=sort_segments)


def write_segments_to_destination(destination, segments, separator="\t", extra="all", extra_separator=";", extra_fill="", sort_segments=True):
    if sort_segments:
        segments = sorted(segments, key=lambda s: (s.chromosome, s.start_position.coordinate, s.end_position.coordinate))
    header_exntries = [CHR, START, END]
    if extra is not None:
        header_exntries.append(EXTRA)
    writer = csv.DictWriter(destination, fieldnames=header_exntries, delimiter=separator)
    writer.writeheader()
    for segment in segments:
        data = {}
        data[CHR] = segment.chromosome
        data[START] = segment.start_position.coordinate
        data[END] = segment.end_position.coordinate
        if extra is not None:
            extra_strings = []
            if extra_description_is_all(extra=extra):
                for key, value in segment.extra.items():
                    extra_strings.append("{extra_name}={extra_value}".format(extra_name=key.lower(), extra_value=value if value is not None else extra_fill))
            else:
                for entry in extra:
                    extra_strings.append("{extra_name}={extra_value}".format(extra_name=str(entry).lower(), extra_value=segment.extra.get(entry, extra_fill)))
            extra_string = extra_separator.join(extra_strings)
            data[EXTRA] = extra_string
        writer.writerow(data)


def write_scnt_to_file(file_name, segments, scnt, clone_ids=None, separator="\t", extra="all", extra_separator=";", extra_fill="", sort_segments=True, inplace=True):
    with open(file_name, "wt") as destination:
        write_scnt_to_destination(destination=destination, scnt=scnt, segments=segments, clone_ids=clone_ids, separator=separator, extra=extra, extra_separator=extra_separator,
                                  extra_fill=extra_fill, sort_segments=sort_segments, inplace=inplace)


def iter_segments_scnt_dummy(segments, scnt, clone_ids=None, inplace=True):
    if clone_ids is None:
        clone_ids = sorted(scnt.keys())
    for segment in segments:
        if not inplace:
            segment = deepcopy(segment)
        if COPY_NUMBER in segment.extra:
            yield segment
            continue
        sid = segment.stable_id_non_hap
        segment.extra[COPY_NUMBER] = {}
        for clone_id in clone_ids:
            clone_specific_entries = {}
            for haplotype in [Haplotype.A, Haplotype.B]:
                clone_specific_entries[haplotype.value] = scnt[clone_id].get_cn(sid=sid, haplotype=haplotype, default=0)
            segment.extra[COPY_NUMBER][clone_id] = dict(clone_specific_entries)
        yield segment


def write_scnt_to_destination(destination, segments, scnt, clone_ids=None, separator="\t", extra="all", extra_separator=";", extra_fill="", sort_segments=True, inplace=True):
    if clone_ids is None:
        clone_ids = sorted(scnt.keys())
    segments = iter_segments_scnt_dummy(segments=segments, scnt=scnt, clone_ids=clone_ids, inplace=inplace)
    if extra is None:
        extra = [COPY_NUMBER]
    elif not extra_description_is_all(extra) and COPY_NUMBER not in extra:
        extra.append(COPY_NUMBER)
    write_segments_to_destination(destination=destination, segments=segments, separator=separator, extra=extra, extra_separator=extra_separator, extra_fill=extra_fill,
                                  sort_segments=sort_segments)


def iter_segments_scnb_dummy(segments, scnb, clone_ids=None, inplace=True):
    if clone_ids is None:
        clone_ids = sorted(scnb.keys())
    for segment in segments:
        if not inplace:
            segment = deepcopy(segment)
        if COPY_NUMBER_BOUNDARIES in segment.extra:
            yield segment
            continue
        sid = segment.stable_id_non_hap
        segment.extra[COPY_NUMBER_BOUNDARIES] = {}
        for clone_id in clone_ids:
            clone_specific_entries = defaultdict(dict)
            for haplotype in [Haplotype.A, Haplotype.B]:
                for boundary_type in [CNBoundaries.LOWER, CNBoundaries.UPPER]:
                    clone_specific_entries[haplotype.value][boundary_type.value] = scnb[clone_id].get_cnb(sid=sid, hap=haplotype, boundary_type=boundary_type)
            segment.extra[COPY_NUMBER_BOUNDARIES][clone_id] = dict(clone_specific_entries)
        yield segment


def write_scnb_to_file(file_name, segments, scnb, clone_ids=None, separator="\t", extra="all", extra_separator=";", extra_fill="", sort_segments=True, inplace=True):
    with open(file_name, "wt") as destination:
        write_scnb_to_destination(destination=destination, segments=segments, scnb=scnb, clone_ids=clone_ids, separator=separator, extra=extra, extra_fill=extra_fill,
                                  sort_segments=sort_segments, inplace=inplace)


def write_scnb_to_destination(destination, segments, scnb, clone_ids=None, separator="\t", extra="all", extra_separator=";", extra_fill="", sort_segments=True, inplace=True):
    if clone_ids is None:
        clone_ids = sorted(scnb.keys())
    segments = iter_segments_scnb_dummy(segments=segments, scnb=scnb, clone_ids=clone_ids, inplace=inplace)
    if extra is None:
        extra = [COPY_NUMBER_BOUNDARIES]
    elif not extra_description_is_all(extra) and COPY_NUMBER_BOUNDARIES not in extra:
        extra.append(COPY_NUMBER_BOUNDARIES)
    write_segments_to_destination(destination=destination, segments=segments, separator=separator, extra=extra, extra_separator=extra_separator, extra_fill=extra_fill,
                                  sort_segments=sort_segments)


EXTERNAL_NA_ID = AID
SVTYPE = "svtype"


def parse_adjacency_extra_cn_string(adjacency_cn_string):
    clone_specific_dict = ast.literal_eval(node_or_string=adjacency_cn_string)
    result = defaultdict(dict)
    for clone_id, data in clone_specific_dict.items():
        for phasing_str in data:
            cn = int(data[phasing_str])
            phasing = Phasing[phasing_str]
            result[clone_id][phasing] = cn
    return result


def parse_adjacency_old_cn_string(old_cn_string):
    clone_specific_data = old_cn_string.split(";")
    result = {}
    for clone_specific_string in clone_specific_data:
        clone_specific_string = clone_specific_string[1:-1]
        clone, cnaa, cnab, cnba, cnbb = clone_specific_string.split(",")
        cnaa, cnab, cnba, cnbb = int(cnaa), int(cnab), int(cnba), int(cnbb)
        result[clone] = {
            Phasing.AA: cnaa,
            Phasing.AB: cnab,
            Phasing.BA: cnba,
            Phasing.BB: cnbb
        }
    return result


def parse_adjacency_extra(extra, extra_separator=";"):
    result = {}
    extra_entries = extra.split(extra_separator)
    for entry in extra_entries:
        key, value = entry.split("=")
        if key == COPY_NUMBER:
            value = parse_adjacency_extra_cn_string(adjacency_cn_string=value)
        result[key] = value
    return result


def read_adjacencies_from_file(file_name, extra_separator=";", separator="\t"):
    with open(file_name, "rt") as source:
        return read_adjacencies_from_source(source=source, extra_separator=extra_separator, separator=separator)


def read_adjacencies_from_source(source, extra_separator=";", separator="\t"):
    return list(stream_adjacencies_from_source(source=source, extra_separator=extra_separator, separator=separator))


def stream_adjacencies_from_source(source, extra_separator=";", separator="\t"):
    reader = csv.DictReader(source, delimiter=separator)
    for row in reader:
        chr1 = row[CHR1]
        chr2 = row[CHR2]
        coord1 = int(row[COORD1])
        coord2 = int(row[COORD2])
        strand1 = Strand.from_pm_string(string=row[STRAND1])
        strand2 = Strand.from_pm_string(string=row[STRAND2])
        extra_dict = {EXTERNAL_NA_ID: row[EXTERNAL_NA_ID]}
        if OLD_ADJACENCIES_CNS in row:
            extra_dict[COPY_NUMBER] = parse_adjacency_old_cn_string(old_cn_string=row[OLD_ADJACENCIES_CNS])
        if EXTRA in row:
            supp_extra_dict = parse_adjacency_extra(extra=row[EXTRA], extra_separator=extra_separator)
            extra_dict.update(supp_extra_dict)
        pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
        pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
        adjacency = Adjacency(position1=pos1, position2=pos2, extra=extra_dict)
        if ADJACENCY_TYPE in extra_dict:
            adjacency.adjacency_type = AdjacencyType.from_name(name=extra_dict[ADJACENCY_TYPE])
            del extra_dict[ADJACENCY_TYPE]
        else:
            adjacency.adjacency_type = AdjacencyType.REFERENCE if extra_dict[EXTERNAL_NA_ID].startswith("r") else AdjacencyType.NOVEL
        yield adjacency


def write_adjacencies_to_file(file_name, adjacencies, extra="all", extra_fill="", extra_separator=";", separator="\t", sort_adjacencies=True):
    if sort_adjacencies:
        adjacencies = sorted(adjacencies, key=lambda a: (a.position1.chromosome, a.position1.coordinate, a.position2.chromosome, a.position2.coordinate))
    with open(file_name, "wt") as destination:
        write_adjacencies_to_destination(destination=destination, adjacencies=adjacencies,
                                         extra=extra, extra_fill=extra_fill, extra_separator=extra_separator, separator=separator,
                                         sort_adjacencies=False)


def write_adjacencies_to_destination(destination, adjacencies, extra="all", extra_fill="", extra_separator=";", separator="\t", sort_adjacencies=True):
    if sort_adjacencies:
        adjacencies = sorted(adjacencies, key=lambda a: (a.position1.chromosome, a.position1.coordinate, a.position2.chromosome, a.position2.coordinate))
    header_entries = [AID, CHR1, COORD1, STRAND1, CHR2, COORD2, STRAND2]
    if extra is not None:
        header_entries.append(EXTRA)
    writer = csv.DictWriter(destination, fieldnames=header_entries, delimiter=separator)
    writer.writeheader()
    for adjacency in adjacencies:
        data = {}
        data[AID] = adjacency.extra.get(EXTERNAL_NA_ID, adjacency.idx)
        data[CHR1] = str(adjacency.position1.chromosome)
        data[COORD1] = str(adjacency.position1.coordinate)
        data[STRAND1] = str(adjacency.position1.strand)
        data[CHR2] = str(adjacency.position2.chromosome)
        data[COORD2] = str(adjacency.position2.coordinate)
        data[STRAND2] = str(adjacency.position2.strand)
        if extra is not None:
            extra_strings = []
            if extra_description_is_all(extra=extra):
                for key, value in adjacency.extra.items():
                    if isinstance(value, (list, tuple)):
                        value = ",".join(map(str, value))
                    extra_strings.append("{extra_name}={extra_value}".format(extra_name=str(key).lower(), extra_value=value if value is not None else extra_fill))
                extra_strings.append("{extra_name}={extra_value}".format(extra_name=ADJACENCY_TYPE.lower(), extra_value=adjacency.adjacency_type.value))
            else:
                for entry in extra:
                    value = adjacency.extra.get(entry, extra_fill)
                    if isinstance(value, (list, tuple)):
                        value = ",".join(map(str, value))
                    extra_strings.append("{extra_name}={extra_value}".format(extra_name=str(entry).lower(), extra_value=value))
            extra_string = extra_separator.join(extra_strings)
            data[EXTRA] = extra_string
        writer.writerow(data)


def suitable_extra_entry(entry, extra="all"):
    if extra_description_is_all(extra=extra):
        return True
    return str(entry).lower() in extra


VCF_SAMPLE_EXTRA_FIELDS = {
    COPY_NUMBER.lower(),
    COPY_NUMBER_BOUNDARIES.lower(),
    VCF_GENOTYPE.lower()
}


def is_info_extra_entry(entry):
    return str(entry).lower() not in VCF_SAMPLE_EXTRA_FIELDS


def is_format_extra_entry(entry):
    return not is_info_extra_entry(entry=entry)


def get_vcf_info_string(adjacency, extra_fields, extra_fill=""):
    result = []
    result.append("CHR2={chromosome}".format(chromosome=adjacency.position2.chromosome))
    result.append("END={coordinate}".format(coordinate=adjacency.position2.coordinate))
    result.append("STRANDS={s1}{s2}".format(s1=str(adjacency.position1.strand), s2=str(adjacency.position2.strand)))
    extra_lower = {key.lower(): value for key, value in adjacency.extra.items() if key.upper() not in {"CHR2", "END", "STRANDS"}}
    for key in extra_fields:
        if key.upper() in {"CHR2", "END", "STRANDS"}:
            continue
        value = extra_lower.get(key.lower(), extra_fill)
        if isinstance(value, (list, tuple)):
            value = ",".join(map(str, value))
        result.append("{entry_id}={value}".format(entry_id=key, value=value))
    return ";".join(result)


def get_vcf_format_string(adjacency, clone_id, format_fields, extra_fill=""):
    result = []
    if COPY_NUMBER in adjacency.extra:
        total_cn = sum([adjacency.extra[COPY_NUMBER][clone_id][ph] for ph in [Phasing.AA, Phasing.AB, Phasing.BA, Phasing.BB]])
        genotype_value = "0/0" if total_cn == 0 else "./."
    else:
        genotype_value = "./."
    extra_lower = {key.lower(): value for key, value in adjacency.extra.items()}
    for entry_id in format_fields:
        if entry_id.lower() == VCF_GENOTYPE.lower():
            value = genotype_value
        elif entry_id.lower() == COPY_NUMBER.lower() and COPY_NUMBER in adjacency.extra:
            value = ",".join([str(adjacency.extra[COPY_NUMBER][clone_id][ph]) for ph in [Phasing.AA, Phasing.AB, Phasing.BA, Phasing.BB]])
        else:
            value = str(extra_lower.get(entry_id.lower(), extra_fill))
        result.append(value)
    return ":".join(result)


class VCFOutputFormat(Enum):
    SNIFFLES = "Sniffles"
    VCFProper = "VCF"


def write_adjacencies_to_vcf_sniffles_file(file_name, adjacencies, extra="all", extra_fill=".", sort_adjacencies=True):
    with open(file_name, "wt") as destination:
        write_adjacencies_to_vcf_sniffles_destination(destination=destination, adjacencies=adjacencies, extra=extra, extra_fill=extra_fill, sort_adjacencies=sort_adjacencies)


def write_adjacencies_to_vcf_sniffles_destination(destination, adjacencies, extra="all", extra_fill=".", sort_adjacencies=True,
                                                  dummy_clone="dummy_clone", clone_suffix=""):
    if sort_adjacencies:
        adjacencies = sorted(adjacencies, key=lambda a: (a.position1.chromosome, a.position1.coordinate, a.position2.chromosome, a.position2.coordinate))
    adjacencies = list(adjacencies)
    today = datetime.datetime.today()
    print("##fileformat=VCFv4.1", file=destination)
    print("##fileDate={year}{month:0>2}{day:0>2}".format(year=today.year, month=today.month, day=today.day), file=destination)
    print("##source=RCK", file=destination)
    chromosomes = set()
    svtypes = set()
    extra_fields = set()
    for adj in adjacencies:
        chromosomes.add(adj.position1.chromosome)
        chromosomes.add(adj.position2.chromosome)
        if SVTYPE in adj.extra:
            svtypes.add(adj.extra[SVTYPE])
        for key in adj.extra.keys():
            if suitable_extra_entry(entry=key, extra=extra):
                extra_fields.add(key)
    chromosomes = sorted(chromosomes)
    svtypes = sorted(svtypes)
    for chromosome in chromosomes:
        print("##contig=<ID={chromosome}>".format(chromosome=chromosome), file=destination)
    for svtype in svtypes:
        print("##ALT=<ID={svtype},Description=\"{descr}\">".format(svtype=svtype.upper(), descr=svtype), file=destination)
    print("##INFO=<ID=CHR2,Number=1,Type=String,Description=\"chromosome for second position of the reported adjacency\">", file=destination)
    print("##INFO=<ID=END,Number=1,Type=Integer,Description=\"coordinate for second position of the reported adjacency\">", file=destination)
    print("##INFO=<ID=STRANDS,Number=1,Type=String,Description=\"strands _x__y_ where _x_ corresponds tot he strand for first position, and _y_ for the \">", file=destination)
    clone_ids = None
    extra_info_fields_and_numbers = {}
    extra_format_fields_and_numbers = {"GT": "1"}
    for adjacency in adjacencies:
        for key, value in adjacency.extra.items():
            if key not in extra_fields:
                continue
            if is_info_extra_entry(entry=key):
                entry_id = str(key).upper()
                entry_number = "." if isinstance(value, (list, tuple)) else "1"
                if entry_id in extra_info_fields_and_numbers and extra_info_fields_and_numbers[entry_id] != entry_number:
                    entry_number = "1"
                extra_info_fields_and_numbers[entry_id] = entry_number
            elif is_format_extra_entry(entry=key):
                entry_id = str(key).upper()
                entry_number = "1"
                extra_format_fields_and_numbers[entry_id] = entry_number
                if isinstance(value, dict):
                    if clone_ids is None:
                        clone_ids = set(value.keys())
                    else:
                        clone_ids &= set(value.keys())
    if clone_ids is None:
        clone_ids = [dummy_clone]
    clone_names = [clone_id + clone_suffix for clone_id in clone_ids]
    clone_ids = sorted(clone_ids)

    for extra_entry_id, extra_entry_number in extra_info_fields_and_numbers.items():
        if extra_entry_id in ["CHR2", "END", "STRANDS"]:
            continue
        print("##INFO=<ID={entry_id},Number={entry_number},Type=String,Description=\"\">".format(entry_id=extra_entry_id, entry_number=extra_entry_number), file=destination)
    for extra_entry_id, extra_entry_number in extra_format_fields_and_numbers.items():
        print("##FORMAT=<ID={entry_id},Number={entry_number},Type=String,Description=\"\">".format(entry_id=extra_entry_id, entry_number=extra_entry_number), file=destination)
    for clone_id in clone_names:
        print("##SAMPLE=<ID={sample_id}>".format(sample_id=clone_id), file=destination)

    header = ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT"]
    for clone_id in clone_names:
        header.append(clone_id)
    print("\t".join(header), file=destination)

    for adjacency in adjacencies:
        print(adjacency.position1.chromosome, end="\t", file=destination)  # CHROM
        print(adjacency.position1.coordinate, end="\t", file=destination)  # POS
        print(adjacency.extra.get(EXTERNAL_NA_ID, adjacency.idx), end="\t", file=destination)  # ID
        print("N", end="\t", file=destination)  # REF
        print("<{svtype}>".format(svtype=adjacency.extra.get(SVTYPE, "BND")), end="\t", file=destination)  # ALT
        print(".", end="\t", file=destination)  # QUAL
        print("PASS", end="\t", file=destination)  # FILTER
        print(get_vcf_info_string(adjacency=adjacency, extra_fields=sorted(extra_info_fields_and_numbers.keys()), extra_fill=extra_fill), end="\t", file=destination)  # INFO
        print(":".join(extra_format_fields_and_numbers.keys()), end="\t", file=destination)  # FORMAT
        for clone_id in clone_ids:
            print(get_vcf_format_string(adjacency=adjacency, clone_id=clone_id, format_fields=sorted(extra_format_fields_and_numbers.keys())), end="\t", file=destination)  # SAMPLE
        print(file=destination)


def write_acnt_to_file(file_name, acnt, adjacencies, clone_ids=None,
                       extra="all", extra_fill="", extra_separator=";", separator="\t", sort_adjacencies=True, output_reference=True, inplace=True,
                       mix_reference_and_novel=False):
    with open(file_name, "wt") as destination:
        write_acnt_to_destination(destination=destination, acnt=acnt, adjacencies=adjacencies, clone_ids=clone_ids, extra=extra, extra_fill=extra_fill,
                                  extra_separator=extra_separator, separator=separator, sort_adjacencies=sort_adjacencies, output_reference=output_reference, inplace=inplace,
                                  mix_reference_and_novel=mix_reference_and_novel)


def iter_adjacencies_acnt_dummy(adjacencies, acnt, clone_ids=None, inplace=True):
    if clone_ids is None:
        clone_ids = sorted(acnt.keys())
    for adjacency in adjacencies:
        if not inplace:
            adjacency = deepcopy(adjacency)
        if COPY_NUMBER in adjacency.extra:
            yield adjacency
            continue
        aid = adjacency.stable_id_non_phased
        adjacency.extra[COPY_NUMBER] = {}
        for clone_id in clone_ids:
            clone_specific_entries = {}
            for phasing in [Phasing.AA, Phasing.AB, Phasing.BA, Phasing.BB]:
                clone_specific_entries[str(phasing)] = acnt[clone_id].get_cn(aid=aid, phasing=phasing, default=0)
            adjacency.extra[COPY_NUMBER][clone_id] = clone_specific_entries
        yield adjacency


def write_acnt_to_destination(destination, acnt, adjacencies, clone_ids=None,
                              extra="all", extra_fill="", extra_separator=";", separator="\t", sort_adjacencies=True, output_reference=True, inplace=False,
                              mix_reference_and_novel=False):
    entries = list(adjacencies)
    if sort_adjacencies:
        entries = sorted(entries, key=lambda a: (a.position1.chromosome, a.position1.coordinate, a.position2.chromosome, a.position2.coordinate))
    if not mix_reference_and_novel:
        ref_adjacencies = list(filter(lambda a: a.adjacency_type == AdjacencyType.REFERENCE, adjacencies))
        nov_adjacencies = list(filter(lambda a: a.adjacency_type == AdjacencyType.NOVEL, adjacencies))
        entries = nov_adjacencies + ref_adjacencies
    if not output_reference:
        entries = [adj for adj in entries if adj.adjacency_type == AdjacencyType.NOVEL]
    if clone_ids is None:
        clone_ids = sorted(acnt.keys())
    entries = iter_adjacencies_acnt_dummy(adjacencies=entries, acnt=acnt, clone_ids=clone_ids, inplace=inplace)
    if extra is None:
        extra = [COPY_NUMBER]
    elif not extra_description_is_all(extra=extra) and COPY_NUMBER not in extra:
        extra.append(COPY_NUMBER)
    write_adjacencies_to_destination(destination=destination, adjacencies=entries,
                                     extra=extra, extra_fill=extra_fill, extra_separator=extra_separator, separator=separator, sort_adjacencies=False)


def read_acnt_from_source(source, clone_ids=None, extra_separator=";", separator="\t", remove_cn_data_from_adj=True):
    adjacencies = read_adjacencies_from_source(source=source, extra_separator=extra_separator, separator=separator)
    for adj in adjacencies:
        if COPY_NUMBER not in adj.extra:
            raise ValueError("Trying to read Adjacency Copy Number tensor from {source}, but the {cn} entry in the {extra} column (or {extra} column is missing al together)"
                             "".format(source=source, cn=COPY_NUMBER, extra=EXTRA))
    acnt = extract_acnt_from_adjacencies(adjacencies=adjacencies, clone_ids=clone_ids)
    if remove_cn_data_from_adj:
        remove_cn_data_from_adjacencies(adjacencies=adjacencies)
    return adjacencies, acnt


def remove_cn_data_from_adjacencies(adjacencies):
    for adj in adjacencies:
        if COPY_NUMBER in adj.extra:
            del adj.extra[COPY_NUMBER]


def extract_acnt_from_adjacencies(adjacencies, clone_ids=None):
    if clone_ids is None:
        clone_ids = set()
        for adj in adjacencies:
            clone_ids.update(set(adj.extra[COPY_NUMBER].keys()))
    clone_ids = sorted(clone_ids)
    acnt = {clone_id: AdjacencyCopyNumberProfile() for clone_id in clone_ids}
    for adj in adjacencies:
        cn_data = adj.extra[COPY_NUMBER]
        aid = adj.stable_id_non_phased
        for clone_id in clone_ids:
            if clone_id not in cn_data:
                raise ValueError("Clone {cid} is not present in CN data for adjacency {adj}".format(cid=clone_id, adj=aid))
            for phasing in cn_data[clone_id]:
                acnp = acnt[clone_id]
                acnp._set_cn_record(aid=aid, phasing=phasing, cn=cn_data[clone_id][phasing])
    return acnt


def read_acnt_from_file(file_name, clone_ids=None, extra_separator=";", separator="\t", remove_cn_data_from_adj=True):
    with open(file_name, "rt") as source:
        return read_acnt_from_source(source=source, clone_ids=clone_ids, extra_separator=extra_separator, separator=separator, remove_cn_data_from_adj=remove_cn_data_from_adj)


def parse_acn_string(string_entry, separator="\t", cn_separator=";"):
    cns = {}
    data = string_entry.split(separator)
    naid = data[0]
    chr1 = data[1]
    coord1 = int(data[2])
    strand1 = Strand.from_pm_string(string=data[3])
    chr2 = data[4]
    coord2 = int(data[5])
    strand2 = Strand.from_pm_string(string=data[6])
    cns_string_entry = data[7]
    p1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
    p2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
    adj_type = AdjacencyType.REFERENCE if "r" in naid else AdjacencyType.NOVEL
    adjacency = Adjacency(position1=p1, position2=p2, adjacency_type=adj_type, extra={EXTERNAL_NA_ID: naid})
    cn_data = cns_string_entry.split(cn_separator)
    for entry in cn_data:
        entry = entry[1:-1]
        clone_id, cnAA, cnAB, cnBA, cnBB = entry.split(",")
        cnAA, cnAB, cnBA, cnBB = int(cnAA), int(cnAB), int(cnBA), int(cnBB)
        cns[clone_id] = {Phasing.AA: cnAA, Phasing.AB: cnAB, Phasing.BA: cnBA, Phasing.BB: cnBB}
    return adjacency, cns


def read_acnt_old(file_name, clone_ids, separator="\t", cn_separator=";"):
    clone_ids = sorted(clone_ids)
    result = {clone_id: AdjacencyCopyNumberProfile() for clone_id in clone_ids}
    adjacencies = []
    with open(file_name, "rt") as source:
        for line_cnt, line in enumerate(source):
            line = line.strip()
            if line_cnt == 0 or len(line) == 0 or line.startswith("#"):
                continue
            adjacency, cns = parse_acn_string(string_entry=line, separator=separator, cn_separator=cn_separator)
            if EXTERNAL_NA_ID in adjacency.extra and adjacency.extra[EXTERNAL_NA_ID].startswith("r"):
                adjacency.adjacency_type = AdjacencyType.REFERENCE
            else:
                adjacency.adjacency_type = AdjacencyType.NOVEL
            aid = adjacency.stable_id_non_phased
            adjacencies.append(adjacency)
            for clone_id in clone_ids:
                if clone_id not in cns:
                    raise Exception()
                for ph in [Phasing.AA, Phasing.AB, Phasing.BA, Phasing.BB]:
                    acnp = result[clone_id]
                    acnp._set_cn_record(aid=aid, phasing=ph, cn=cns[clone_id][ph])
    return adjacencies, result


def get_all_clone_ids_from_dasck_scnt_file(file_name, separator="\t", cn_separator=";", allow_unit_segments=False):
    clone_ids = set()
    with open(file_name, "rt") as source:
        for line_cnt, line in enumerate(source):
            line = line.strip()
            if len(line) == 0 or line.startswith("#") or line_cnt == 0:
                continue
            segment, cns = parse_scn_record(scn_string=line, separator=separator, cn_separator=cn_separator, allow_unit_segments=allow_unit_segments)
            for clone_id in cns.keys():
                clone_ids.add(clone_id)
    return sorted(clone_ids)


def parser_position_string_entry(position_string, separator):
    data = position_string.split(separator)
    chr_name = data[0]
    coordinate = int(data[1])
    strand = Strand.from_pm_string(string=data[2])
    return Position(chromosome=chr_name, coordinate=coordinate, strand=strand)


def read_positions(file_name, separator="\t"):
    result = []
    with open(file_name, "rt") as source:
        for line_cnt, line in enumerate(source):
            line = line.strip()
            if len(line) == 0 or line.startswith("#") or line_cnt == 0:
                continue
            position = parser_position_string_entry(position_string=line, separator=separator)
            result.append(position)
    return result


def read_positions_from_file(file_name, separator="\t", extra_separator=";"):
    with open(file_name, "rt") as source:
        return read_positions_from_source(source=source, separator=separator, extra_separator=extra_separator)


def read_positions_from_source(source, separator="\t", extra_separator=";"):
    return list(stream_positions_from_source(source=source, separator=separator, extra_separator=extra_separator))


def stream_positions_from_source(source, separator="\t", extra_separator=";"):
    reader = csv.DictReader(source, delimiter=separator)
    for row in reader:
        chromosome = row[CHR]
        coordinate = int(row[COORD])
        strand = Strand.from_pm_string(string=row[STRAND])
        yield Position(chromosome=chromosome, coordinate=coordinate, strand=strand)


def read_adjacency_groups_from_file(file_name, default_group_type=AdjacencyGroupType.MOLECULE, separator="\t", aids_separator=",", extra_separator=";"):
    with open(file_name, "rt") as source:
        return read_adjacency_groups_from_source(source=source, default_group_type=default_group_type, separator=separator,
                                                 aids_separator=aids_separator, extra_separator=extra_separator)


def read_adjacency_groups_from_source(source, default_group_type=AdjacencyGroupType.MOLECULE, separator="\t", aids_separator=",", extra_separator=";"):
    return list(stream_adjacency_groups_from_source(source=source, default_group_type=default_group_type, separator=separator,
                                                    aids_separator=aids_separator, extra_separator=extra_separator))


def stream_adjacency_groups_from_source(source, default_group_type=AdjacencyGroupType.MOLECULE, separator="\t", aids_separator=",", extra_separator=";"):
    reader = csv.DictReader(source, delimiter=separator)
    for row in reader:
        gid = row[GID]
        aids = row[AIDS].split(aids_separator)
        extra_dict = {}
        if EXTRA in row:
            extra_dict = parse_adj_groups_extra(extra_string=row[EXTRA], extra_separator=extra_separator)
        group_type = extra_dict.get(AG_TYPE, default_group_type)
        yield AdjacencyGroup(gid=gid, aids=aids, group_type=group_type, extra=extra_dict)


def write_positions_to_file(file_name, positions, extra="all", extra_fill="", separator="\t", extra_separator=";", sort=True):
    with open(file_name, "wt") as destination:
        write_positions_to_destinations(destination=destination, positions=positions, extra=extra, extra_fill=extra_fill, separator=separator, extra_separator=extra_separator,
                                        sort=sort)


def write_positions_to_destinations(destination, positions, extra="all", extra_fill="", separator="\t", extra_separator=";", sort=True):
    header_entries = [CHR, COORD, STRAND]
    if sort:
        positions = sorted(positions, key=lambda p: (p.chromosome, p.coordinate, p.strand))
    writer = csv.DictWriter(destination, fieldnames=header_entries, delimiter=separator)
    writer.writeheader()
    for position in positions:
        data = {}
        data[CHR] = position.chromosome
        data[COORD] = position.coordinate
        data[STRAND] = position.strand
        writer.writerow(data)


def parse_adj_group_extra_labeling_string(extra_labeling_string):
    return list(map(int, extra_labeling_string.split(",")))


def parse_adj_groups_extra(extra_string, extra_separator=";"):
    data = extra_string.split(extra_separator)
    result = {}
    for entry in data:
        key, value = entry.split("=")
        key = key.lower()
        if key == AG_TYPE:
            value = AdjacencyGroupType.from_string(string=value)
        elif key == FALSE_POSITIVE:
            value = float(value)
        elif key == AG_LABELING:
            value = list(map(int, value.split(",")))
        result[key] = value
    return result


def write_adjacency_groups_to_file(file_name, adjacency_groups, separator="\t", aids_separator=",",
                                   extra="all", extra_separator=";", extra_fill=""):
    with open(file_name, "wt") as destination:
        write_adjacency_groups_to_destination(destination=destination, adjacency_groups=adjacency_groups, separator=separator,
                                              aids_separator=aids_separator, extra=extra, extra_separator=extra_separator, extra_fill=extra_fill)


def write_adjacency_groups_to_destination(destination, adjacency_groups, separator="\t", aids_separator=",",
                                          extra="all", extra_separator=";", extra_fill=""):
    header_entries = [GID, AIDS, EXTRA]
    if extra is None:
        extra = [AG_TYPE]
    # if extra is not None:
    #     header_entries.append(EXTRA)
    writer = csv.DictWriter(destination, fieldnames=header_entries, delimiter=separator)
    writer.writeheader()
    for ag in adjacency_groups:
        data = {}
        data[GID] = ag.gid
        data[AIDS] = aids_separator.join(map(str, ag.adjacencies_ids))
        extra_strings = ["{ag_type_string}={ag_type}".format(ag_type_string=AG_TYPE, ag_type=ag.group_type.value)]
        if extra_description_is_all(extra=extra):
            for key, value in ag.extra.items():
                if isinstance(value, (list, tuple)):
                    value = ",".join(map(str, value))
                elif isinstance(value, AdjacencyGroupType):
                    value = value.value
                extra_strings.append("{extra_name}={extra_value}".format(extra_name=str(key).lower(), extra_value=value if value is not None else extra_fill))
        else:
            for entry in extra:
                if entry == AG_TYPE:
                    continue
                value = ag.extra.get(entry, extra_fill)
                if isinstance(value, (list, tuple)):
                    value = ",".join(map(str, value))
                extra_strings.append("{extra_name}={extra_value}".format(extra_name=str(entry).lower(), extra_value=value))
        agt_present = False
        for extra_string in extra_strings:
            agt_present |= "{ag_type_string}=".format(ag_type_string=AG_TYPE) in extra_string
        if not agt_present:
            extra_strings.append("{ag_type_string}={ag_type}".format(ag_type_string=AG_TYPE, ag_type=ag.group_type.value))
        extra_string = extra_separator.join(extra_strings)
        data[EXTRA] = extra_string
        writer.writerow(data)


def get_segment_string_entry(segment, separator="\t"):
    chr_name_entry = "{chr_name}".format(chr_name=str(segment.start_position.chromosome))
    start_entry = "{start_coord}".format(start_coord=str(segment.start_position.coordinate))
    end_entry = "{end_coord}".format(end_coord=str(segment.end_position.coordinate))
    return separator.join([chr_name_entry, start_entry, end_entry])


def write_segments(file_name, segments, separator="\t"):
    with open(file_name, "wt") as destination:
        segments = sorted(segments, key=lambda s: (s.start_position.chromosome, s.start_position.coordinate, s.end_position.coordinate))
        print("chr", "start", "end", sep=separator, file=destination)
        for segment in segments:
            string_entry = get_segment_string_entry(segment=segment, separator=separator)
            print(string_entry, file=destination)


def parse_segment_string_entry(string_entry, separator="\t", allow_unit_segments=False):
    data = string_entry.split(separator)
    chr_name = data[0]
    start_corodinate = int(data[1])
    end_coordinate = int(data[2])
    sp = Position(chromosome=chr_name, coordinate=start_corodinate, strand=Strand.REVERSE)
    ep = Position(chromosome=chr_name, coordinate=end_coordinate, strand=Strand.FORWARD)
    return Segment(start_position=sp, end_position=ep, allow_unit_length=allow_unit_segments)


def read_segments(file_name, separator="\t", allow_unit_segments=False):
    result = []
    with open(file_name, "rt") as source:
        for line_cnt, line in enumerate(source):
            line = line.strip()
            if line_cnt == 0 or len(line) == 0 or line.startswith("#"):
                continue
            segment = parse_segment_string_entry(string_entry=line, separator=separator, allow_unit_segments=allow_unit_segments)
            result.append(segment)
    return result


def get_cn_boundaries_string_entry(segment, scn_boundaries, clone_ids, cn_separator=";"):
    result = []
    sid = segment.stable_id_non_hap
    for clone_id in clone_ids:
        lower_a, upper_a = scn_boundaries[clone_id][sid][Haplotype.A]
        lower_b, upper_b = scn_boundaries[clone_id][sid][Haplotype.B]
        string_a = "{lower}-{upper}".format(lower=lower_a, upper=upper_a)
        string_b = "{lower}-{upper}".format(lower=lower_b, upper=upper_b)
        string_entry = "({clone_id},{boundaries_a},{boundaries_b})".format(clone_id=str(clone_id), boundaries_a=str(string_a), boundaries_b=str(string_b))
        result.append(string_entry)
    return cn_separator.join(result)


def write_scn_boundaries(file_name, segments, scn_boundaries, separator="\t", cn_separator=";"):
    clone_ids = sorted(scn_boundaries.keys())
    with open(file_name, "wt") as destination:
        print("chr", "start", "end", "CNBoundaries", sep=separator, file=destination)
        for s in segments:
            string_entries = []
            string_entries.append("{chr_name}".format(chr_name=str(s.start_position.chromosome)))
            string_entries.append("{start}".format(start=str(s.start_position.coordinate)))
            string_entries.append("{end}".format(end=str(s.end_position.coordinate)))
            string_entries.append(get_cn_boundaries_string_entry(segment=s, scn_boundaries=scn_boundaries, clone_ids=clone_ids, cn_separator=cn_separator))
            string_entry = separator.join(string_entries)
            print(string_entry, file=destination)


def parse_cn_boundaries_entry_string(string_entry):
    data = string_entry.split("-")
    return int(data[0]), int(data[1])


def parse_cn_boundaries_string(string_entry, cn_separator=";"):
    result = {}
    data = string_entry.split(cn_separator)
    for entry in data:
        entry = entry[1:-1]
        entry_data = entry.split(",")
        clone_id = entry_data[0]
        boundaries_a_string = entry_data[1]
        boundaries_b_string = entry_data[2]
        boundaries_a = parse_cn_boundaries_entry_string(string_entry=boundaries_a_string)
        boundaries_b = parse_cn_boundaries_entry_string(string_entry=boundaries_b_string)
        result[clone_id] = {Haplotype.A: boundaries_a, Haplotype.B: boundaries_b}
    return result


def parse_scn_boundaries_string_entry(string_entry, separator="\t", cn_separator=";", allow_unit_segments=False):
    data = string_entry.split(separator)
    chr_name = data[0]
    start_coordinate = int(data[1])
    end_coordinate = int(data[2])
    cn_boundaries_string = data[3]
    sp = Position(chromosome=chr_name, coordinate=start_coordinate, strand=Strand.REVERSE)
    ep = Position(chromosome=chr_name, coordinate=end_coordinate, strand=Strand.FORWARD)
    segment = Segment(start_position=sp, end_position=ep, allow_unit_length=allow_unit_segments)
    boundaries = parse_cn_boundaries_string(string_entry=cn_boundaries_string, cn_separator=cn_separator)
    return segment, boundaries


def read_scn_boundaries(file_name, segments, clone_ids, separator="\t", cn_separator=";", allow_unit_segments=False):
    result = {clone_id: {s.stable_id_non_hap: {Haplotype.A: None, Haplotype.B: None} for s in segments} for clone_id in clone_ids}
    with open(file_name, "rt") as source:
        for line_cnt, line in enumerate(source):
            line = line.strip()
            if line_cnt == 0 or len(line) == 0 or line.startswith("#"):
                continue
            segment, boundaries = parse_scn_boundaries_string_entry(string_entry=line, separator=separator, cn_separator=cn_separator, allow_unit_segments=allow_unit_segments)
            for clone_id in clone_ids:
                result[clone_id][segment.stable_id_non_hap][Haplotype.A] = boundaries[clone_id][Haplotype.A]
                result[clone_id][segment.stable_id_non_hap][Haplotype.B] = boundaries[clone_id][Haplotype.B]
    return result


def get_full_path(path):
    return os.path.abspath(os.path.expanduser(path))



import argparse
import os

import sys
from copy import deepcopy

current_file_level = 2
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

import rck
from rck.core.io import read_adjacencies_from_file, \
    get_logging_cli_parser, get_standard_logger_from_args, get_full_path, read_scnt_from_file, read_positions_from_source, \
    write_segments_to_file, write_scnt_to_file
from rck.core.structures import refined_scnt, refined_scnt_with_adjacencies_and_telomeres


def main():
    parser = argparse.ArgumentParser(prog="RCK-UTILS-input-refine", parents=[get_logging_cli_parser()])
    parser.add_argument("--version", action="version", version=rck.version)
    parser.add_argument("--scnt", required=True)
    parser.add_argument("--adjacencies", required=True)
    parser.add_argument("--clone-ids", default=None)
    parser.add_argument("--scnt-separator", default="\t")
    parser.add_argument("--adjacencies-separator", default="\t")
    parser.add_argument("--no-merge-fragments", action="store_false", dest="merge_fragments")
    parser.add_argument("--fragments-max-merge-gap", type=int, default=1000000000)
    parser.add_argument("--no-fill-gaps-fragments", action="store_false", dest="fill_gaps_fragments")
    parser.add_argument("--fragments-max-fill-gap", type=int, default=1000000000)
    parser.add_argument("--no-allow-unit-segments", action="store_false", dest="allow_unit_segments")
    parser.add_argument("--telomere-positions", type=argparse.FileType("rt"))
    parser.add_argument("--telomere-positions-separator", default="\t")
    parser.add_argument("--output-scnt", required=True)
    parser.add_argument("--output-fragments", required=True)
    args = parser.parse_args()
    logger = get_standard_logger_from_args(args=args, program_name="RCK-UTILS-input-refine")
    clone_ids = args.clone_ids.split(",") if args.clone_ids is not None else None
    scnt_file = get_full_path(args.scnt_file)
    adj_file = get_full_path(args.adj)
    segments, scnt = read_scnt_from_file(file_name=scnt_file, clone_ids=clone_ids, separator=args.scnt_separator)
    clone_ids = sorted(set(scnt.keys()))
    segments, scnt, segments_ids_mapping = refined_scnt(segments=segments, scnt=scnt,
                                                        merge_fragments=args.merge_fragments, max_merge_gap=args.fragments_max_merge_gap,
                                                        fill_gaps=args.fill_gaps_fragments, max_fill_gap=args.fragments_max_fill_gap)

    adjacencies = read_adjacencies_from_file(file_name=adj_file, separator=args.adjacencies_separator)
    if args.telomere_positions is not None:
        telomere_positions = read_positions_from_source(source=args.telomere_positions, separator=args.telomere_positions_separator)
    else:
        telomere_positions = []
    fragments = deepcopy(segments)
    segments, scnt = refined_scnt_with_adjacencies_and_telomeres(segments=segments, scnt=scnt, adjacencies=adjacencies, telomere_positions=telomere_positions)
    refined_scnt_file = os.path.expanduser(args.refined_scnt_file)
    refined_scnt_file = os.path.abspath(refined_scnt_file)
    fragments_file = get_full_path(path=args.output_fragments)

    write_segments_to_file(file_name=fragments_file, segments=fragments)
    write_scnt_to_file(file_name=refined_scnt_file, scnt=scnt, segments=segments)


if __name__ == "__main__":
    main()



import argparse
import itertools
import random
import re

import sys
import os
import math
from collections import Counter
import pandas as pd

current_file_level = 3
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

import rck
from rck.core.io import write_adjacencies_to_destination, read_adjacencies_from_source
from rck.utils.adj.convert import *
from rck.utils.adj.process import filter_adjacencies_by_chromosomal_regions, get_shared_nas_parser, ORIGIN_IDS


def get_str_size_label(int_size):
    if int_size < 1000:
        return str(int_size)
    elif int_size < 1000000:
        return str(math.floor(int_size / 1000)) + "KB"
    return str(math.floor(int_size / 1000000)) + "MB"


def add_bar_values(ax, bars):
    for bar in bars:
        height = bar.get_height()
        ax.text(bar.get_x() + bar.get_width() / 2., 1.01 * height,
                '%d' % int(height),
                ha='center', va='bottom', fontsize=20)


def get_length(na, treat_ins_separately=True):
    if treat_ins_separately and na.extra.get(SVTYPE, "").lower() == "ins":
        result = int(na.extra.get(SVLEN, -1))
        if result > -1:
            return result
    return na.distance_non_hap


def main():
    parser = argparse.ArgumentParser(prog="RCK-UTILS-ADJ-STATS")
    parser.add_argument('--version', action='version', version=rck.version)
    ######
    shared_parser = get_shared_nas_parser()
    shared_parser.add_argument("--no-vis", action="store_false", dest="vis")
    shared_parser.add_argument("--vis-interactive", action="store_true")
    shared_parser.add_argument("--output-dir", "-o", dest="output_dir", default=os.getcwd())
    ######
    subparsers = parser.add_subparsers(title="command", dest="command")
    subparsers.required = True
    ######
    cnt_parser = subparsers.add_parser("cnt", parents=[shared_parser], help="Counting stats for RCK NAS in input file")
    cnt_parser.add_argument("rck_nas", type=argparse.FileType("rt"), default=sys.stdin)
    cnt_parser.add_argument("--ann", action="store_true")
    cnt_parser.add_argument("--ann-list", nargs=1)
    cnt_parser.add_argument("--ann-field", default="svtype", type=str)
    cnt_parser.add_argument("--ann-missing", default="unknown")
    cnt_parser.add_argument("--bins", nargs=1,
                            default=",".join(["-1", "0", "50", "100", "500", "1000", "5000", "10000", "50000",
                                              "100000", "500000", "1000000", "5000000", "10000000"]),
                            # default=",".join(["-1", "0", "50", "100", "200", "300", "400", "500"])
                            )
    cnt_parser.add_argument("--cnt-output-subdir", default="cnt")
    cnt_parser.add_argument("--title", default="")
    cnt_parser.add_argument("--no-bar-values", action="store_false", dest="bar_values")
    # cnt_parser.add_argument("--per-chr", action="store_true")
    ######
    lr_parser = subparsers.add_parser("lr", parents=[shared_parser], help="Counting stats w.r.t. long read information in the input RCK NAS file")
    lr_parser.add_argument("rck_nas", type=argparse.FileType("rt"), default=sys.stdin)
    lr_parser.add_argument("--lr-field", default="support_read_names")
    lr_parser.add_argument("--title", default="")
    lr_parser.add_argument("--no-bar-values", action="store_false", dest="bar_values")
    lr_parser.add_argument("--lr-output-subdir", default="lr")
    ######
    merged_parser = subparsers.add_parser("merged", parents=[shared_parser], help="Counting statistics over merged RCK NAS")
    merged_parser.add_argument("rck_nas", type=argparse.FileType("rt"), default=sys.stdin)
    merged_parser.add_argument("--origin-field", default=ORIGIN_IDS)
    merged_parser.add_argument("--origin-sep", default=",")
    merged_parser.add_argument("--origin-regex", default=".*_(?P<source>.*)")
    merged_parser.add_argument("--no-origin-field", choices=["skip", "self"])
    merged_parser.add_argument("---merged-output-subdir", default="merged")
    merged_parser.add_argument("--title", default="")
    #######
    support_parser = subparsers.add_parser("support", parents=[shared_parser], help="Counting support statistics over merged RCK NAS")
    support_parser.add_argument("rck_nas", type=argparse.FileType("rt"), default=sys.stdin)
    support_parser.add_argument("--sources", type=argparse.FileType("rt"), nargs="+")
    support_parser.add_argument("--title", default="")
    support_parser.add_argument("--no-bar-values", action="store_false", dest="bar_values")
    support_parser.add_argument("--support-output-subdir", default="support")
    #######
    args = parser.parse_args()
    if args.vis:
        import seaborn as sns
        import matplotlib.pyplot as plt
        sns.set(color_codes=True)
    if args.command == "cnt":
        nas = read_adjacencies_from_source(source=args.rck_nas)
        use_annotations = args.ann
        bins = set()
        for bin_str in args.bins.split(","):
            bin_value = int(bin_str)
            bins.add(bin_value)
        bins = sorted(bins)
        if bins[-1] < 500000000:
            bins.append(500000000)
        if use_annotations:
            annotations = set()
            if len(args.ann_list) == 1 and args.ann_list[0] == "all":
                allow_all_annotations = True
            else:
                allow_all_annotations = False
            for anns_str in args.ann_list:
                anns = [ann.lower() for ann in anns_str.split(",")]
                for ann in anns:
                    annotations.add(ann)
            annotations = sorted(annotations)
            nas_by_anns = defaultdict(list)
            for na in nas:
                na_ann = na.extra.get(args.ann_field, args.ann_missing).lower()
                if allow_all_annotations or na_ann in annotations:
                    nas_by_anns[na_ann].append(na)
        else:
            nas_by_anns = defaultdict(list)
            for na in nas:
                na_ann = str(na.position1.strand) + str(na.position2.strand)
                nas_by_anns[na_ann].append(na)
            # separate figures
        for na_ann, nas in nas_by_anns.items():
            lengths = [get_length(na=na) for na in nas]
            bin_cnts = defaultdict(int)
            for length in lengths:
                for l, r in zip(bins[:-1], bins[1:]):
                    if length < r:
                        bin_cnts[l] += 1
                        break

            if args.vis:
                values = [bin_cnts[i] for i in bins]
                x_axis_values = [i for i in range(len(bins))]
                plt.figure(figsize=(20, 10))
                bars = plt.bar(x_axis_values, values, label=na_ann, color="g")
                if args.bar_values:
                    add_bar_values(ax=plt.gca(), bars=bars)
                x_axis_values = [i for i in range(len(bins))]
                plt.xticks(x_axis_values, ["[{l}-\n{r})".format(l=get_str_size_label(int_size=l), r=get_str_size_label(int_size=r)) for l, r in zip(bins[:-1], bins[1:])])
                plt.tick_params(labelsize=20)
                plt.legend(prop={'size': 20})
                plt.xlabel("{ann} lengths".format(ann=na_ann), fontsize=20)
                plt.ylabel("# of {ann}".format(ann=na_ann), fontsize=20)
                plt.title(args.title, fontsize=20)
                # plt.xticks(bins, [str(b) for b in bins])
                plot_path = os.path.abspath(os.path.join(args.output_dir, args.cnt_output_subdir))
                if not os.path.exists(plot_path):
                    os.makedirs(plot_path)
                plot_file_name = os.path.join(plot_path, "{prefix}_cnt.png".format(prefix=na_ann.replace(os.path.sep, "_")))
                plt.savefig(plot_file_name)
                if args.vis_interactive:
                    plt.show()
                plt.clf()
        # one figure
        fig, ax = plt.subplots()
        fig.set_figheight(20)
        fig.set_figwidth(20)
        diff = .9 / len(list(nas_by_anns.keys()))
        if len(list(nas_by_anns.keys())) % 2 == 0:
            extra_diff = diff / 2
        else:
            extra_diff = 0
        x_axis_values = [i for i in range(len(bins))]
        colors = ["b", "g", "r", "c", "m", "y", "b", "w"]
        for cnt, (na_ann, nas) in enumerate(nas_by_anns.items()):
            lengths = [get_length(na=na) for na in nas]
            bin_cnts = defaultdict(int)
            for length in lengths:
                for l, r in zip(bins[:-1], bins[1:]):
                    if length < r:
                        bin_cnts[l] += 1
                        break
            if args.vis:
                values = [bin_cnts[i] for i in bins]
                x_axis_values_tmp = [x + diff * cnt for x in x_axis_values]
                color = colors[cnt]
                ax.bar(x_axis_values_tmp, values, width=diff, color=color, label="{ann}".format(ann=na_ann))
        plt.xticks(x_axis_values, ["[{l}-\n{r})".format(l=get_str_size_label(int_size=l), r=get_str_size_label(int_size=r)) for l, r in zip(bins[:-1], bins[1:])])
        plt.tick_params(labelsize=20)
        plt.legend(prop={'size': 20})
        plt.xlabel("lenghts", fontsize=20)
        plt.ylabel("# of SVs", fontsize=20)
        plt.title(args.title, fontsize=20)
        plot_path = os.path.abspath(os.path.join(args.output_dir, args.cnt_output_subdir))
        if not os.path.exists(plot_path):
            os.makedirs(plot_path)
        plot_file_name = os.path.join(plot_path, "bins_cnt.png")
        plt.savefig(plot_file_name)
        if args.vis_interactive:
            plt.show()
        plt.gcf()

        fig, ax = plt.subplots()
        fig.set_figheight(20)
        fig.set_figwidth(20)
        x_axis_values = {ann: cnt for cnt, ann in enumerate(nas_by_anns.keys())}
        values = {ann: len(nas_by_anns[ann]) for ann in nas_by_anns.keys()}
        x_axis_values = [x_axis_values[ann] for ann in sorted(nas_by_anns.keys())]
        values = [values[ann] for ann in sorted(nas_by_anns.keys())]
        x_ticks = [ann for ann in sorted(nas_by_anns.keys())]
        bars = plt.bar(x_axis_values, values, color="g")
        if args.bar_values:
            add_bar_values(ax=ax, bars=bars)
        plt.xticks(x_axis_values, x_ticks)
        plt.tick_params(labelsize=20)
        plt.xlabel("SVs", fontsize=20)
        plt.ylabel("# of SVs", fontsize=20)
        plt.title(args.title, fontsize=20)
        plot_path = os.path.abspath(os.path.join(args.output_dir, args.cnt_output_subdir))
        if not os.path.exists(plot_path):
            os.makedirs(plot_path)
        plot_file_name = os.path.join(plot_path, "cnt.png")
        plt.savefig(plot_file_name)
        if args.vis_interactive:
            plt.show()
        plt.gcf()
    elif args.command == "lr":
        nas = read_adjacencies_from_source(source=args.rck_nas)
        reads_to_nas = defaultdict(list)
        for na in nas:
            reads_str = na.extra.get(args.lr_field, "")
            reads = reads_str.split(",")
            for read in reads:
                if len(read) == 0:
                    continue
                reads_to_nas[read].append(na)
        nas_cnts = {len(reads_to_nas[read]) for read in reads_to_nas.keys()}
        x_axis_values = [i for i in range(2, max(nas_cnts))]
        values = defaultdict(int)
        for read in reads_to_nas.keys():
            values[len(reads_to_nas[read])] += 1
        values = [values[i] for i in x_axis_values]
        x_ticks = sorted(x_axis_values)
        fig, ax = plt.subplots()
        fig.set_figheight(20)
        fig.set_figwidth(20)
        bars = plt.bar(x_axis_values, values, color="g")
        if args.bar_values:
            add_bar_values(ax=ax, bars=bars)
        plt.xticks(x_axis_values, x_ticks)
        plt.xlabel("# of SVs")
        plt.ylabel("# of reads supporting x SVs")
        plt.title(args.title)
        plot_path = os.path.abspath(os.path.join(args.output_dir, args.lr_output_subdir))
        if not os.path.exists(plot_path):
            os.makedirs(plot_path)
        plot_file_name = os.path.join(plot_path, "lr_cnt.png")
        plt.savefig(plot_file_name)
        if args.vis_interactive:
            plt.show()
        plt.gcf()
    elif args.command == "merged":
        nas = read_adjacencies_from_source(source=args.rck_nas)
        source_pattern = re.compile(args.origin_regex)
        source_groups_cnt = defaultdict(int)
        for na in nas:
            naid = na.extra.get(EXTERNAL_NA_ID, na.idx)
            if args.origin_field not in na.extra:
                if args.no_origin_field == "skip":
                    continue
                elif args.no_origin_field == "self":
                    origin = naid
                else:
                    raise Exception("Unknown strategy {no_origin_field} for Adjacency {naid} that misses the origin field"
                                    "".format(no_origin_field=args.no_origin_field, naid=naid))
            else:
                origin = na.extra[args.origin_field]
            origin_strings = origin.split(args.origin_sep)
            origins = []
            for origin_string in origin_strings:
                origin_match = source_pattern.match(origin_string)
                if origin_match is None:
                    continue
                origins.append(origin_match.group("source"))
            if len(origins) == 0:
                continue
            origins = tuple(sorted(origins))
            source_groups_cnt[origins] += 1
        source_group_set_cnt = defaultdict(int)
        groups = set()
        pairwise_cnts = defaultdict(int)
        print("\n----- Quantitative subgroups ----\n")
        for key, value in sorted(source_groups_cnt.items(), key=lambda entry: entry[1]):
            for element in key:
                groups.add(element)
            counter = Counter(key)
            if len(set(key)) == 1:
                group = set(key).pop()
                pairwise_cnts[(group, group)] += value
            else:
                for g1, g2 in itertools.combinations(sorted(set(key)), r=2):
                    pairwise_cnts[tuple(sorted([g1, g2]))] += value
                    pairwise_cnts[tuple(reversed(sorted([g1, g2])))] += value
            source_group_set_cnt[tuple(sorted(counter.keys()))] += value
            counter_str = ", ".join(["{name} ({cnt})".format(name=name, cnt=cnt) for name, cnt in sorted(counter.items(), key=lambda entry: (entry[1], entry[0]))])
            print("{key} :: {value}".format(key=counter_str, value=value))

        print("\n----- Groups ----- \n")
        print(", ".join(sorted(groups)))

        print("\n----- By group size (then by size)------\n")
        for key, value in sorted(source_group_set_cnt.items(), key=lambda entry: (len(entry[0]), entry[1])):
            print("{key} :: {value}".format(key=",".join(key), value=value))

        print("\n----- By size (then by group size)------\n")
        for key, value in sorted(source_group_set_cnt.items(), key=lambda entry: (entry[1], len(entry[0]))):
            print("{key} :: {value}".format(key=",".join(key), value=value))

        print("\n----- Pairsize ------\n")
        for key, value in sorted(pairwise_cnts.items()):
            if key[0] >= key[1]:
                print("{key} :: {value}".format(key=",".join(key), value=value))

        if args.vis:
            ser = pd.Series(list(pairwise_cnts.values()),
                            index=pd.MultiIndex.from_tuples(pairwise_cnts.keys()))
            df = ser.unstack().fillna(0)
            df = df.astype('int64')
            sns.set(font_scale=1.5)
            sns.heatmap(df, annot=True, fmt="d")
            plt.title(args.title)
            plot_path = os.path.abspath(os.path.join(args.output_dir, args.merged_output_subdir))
            if not os.path.exists(plot_path):
                os.makedirs(plot_path)
            plot_file_name = os.path.join(plot_path, "pairwise_merged_cnts.png")

            plt.savefig(plot_file_name, bbox_inches='tight')
        if args.vis_interactive:
            plt.show()
    elif args.command == "support":
        max_support = 30
        import statistics
        nas = read_adjacencies_from_source(source=args.rck_nas)
        source_nas = []
        if args.sources is not None:
            for source in args.sources:
                source_nas.extend(read_adjacencies_from_source(source=source))
        source_nas_by_ids = {na.extra.get(EXTERNAL_NA_ID, na.idx): na for na in source_nas}
        counts = defaultdict(int)
        nas_read_cnt = defaultdict(int)
        for na in nas:
            if "support_read_names" in na.extra:
                support_cnt = len(na.extra["support_read_names"].split(","))
                if support_cnt > max_support:
                    support_cnt = max_support
                counts[support_cnt] += 1
                nas_read_cnt[na] += support_cnt
            elif "origin_ids" in na.extra:
                origin_nas = [source_nas_by_ids[origin_id] for origin_id in na.extra["origin_ids"].split(",") if origin_id in source_nas_by_ids]
                if len(origin_nas) == 0:
                    continue
                supports = []
                for origin_na in origin_nas:
                    if "support_read_names" in origin_na.extra:
                        supports.append(len(origin_na.extra["support_read_names"].split(",")))
                support_cnt = int(statistics.mean(supports))
                if support_cnt > max_support:
                    support_cnt = max_support
                counts[support_cnt] += 1
                nas_read_cnt[na] += support_cnt
            else:
                counts[-1] += 1
        x_axis_values = [i for i in range(1, max(counts) + 1)]
        values = [counts[i] for i in x_axis_values]
        x_ticks = sorted(x_axis_values)
        fig, ax = plt.subplots()
        fig.set_figheight(20)
        fig.set_figwidth(20)
        bars = plt.bar(x_axis_values, values, color="g")
        if args.bar_values:
            add_bar_values(ax=ax, bars=bars)
        plt.xticks(x_axis_values, x_ticks)
        plt.xlabel("# of supporting reads")
        plt.ylabel("# of SVs")
        plt.title(args.title)
        plot_path = os.path.abspath(os.path.join(args.output_dir, args.support_output_subdir))
        if not os.path.exists(plot_path):
            os.makedirs(plot_path)
        plot_file_name = os.path.join(plot_path, "support_cnt.png")
        plt.savefig(plot_file_name)
        if args.vis_interactive:
            plt.show()
        plt.gcf()
        top_svs = {na: cnt for na, cnt in nas_read_cnt.items() if cnt >= 28 and na.distance_non_hap > 300}
        result = random.sample(top_svs.keys(), 10)
        for entry in result:
            print(entry, entry.distance_non_hap)


if __name__ == "__main__":
    main()



from enum import Enum

from rck.core.io import EXTERNAL_NA_ID


class ClusteringStrategy(Enum):
    IterativeSlider = 0


def positions_aligned(segments_positions, other_positions):
    segments_positions_ids = {p.stable_id_non_hap for p in segments_positions}
    other_positions_ids = {p.stable_id_non_hap for p in other_positions}
    result = other_positions_ids <= segments_positions_ids
    return result


def adj_groups_concur(adj_groups, adjacencies):
    nas_ids = {na.extra.get(EXTERNAL_NA_ID, na.idx) for na in adjacencies}
    for group in adj_groups:
        for aid in group.adjacencies_ids:
            if aid not in nas_ids:
                return False
    return True
